/*
 * uvector.c.tmpl - uniform vector support code template
 *
 *   Copyright (c) 1999-2004 Shiro Kawai, All rights reserved.
 * 
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the authors nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  $Id: uvector.c.tmpl,v 1.9 2006/11/04 10:04:56 shirok Exp $
 */

#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <string.h>  /* for memcpy() */
#include <gauche.h>
#include <gauche/extend.h>
#include <gauche/builtin-syms.h>
#include "gauche/uvector.h"
#include "gauche/arith.h"
#include "gauche/scmconst.h"

#include "uvectorP.h"

static ScmClass *uvector_cpl[] = {
    SCM_CLASS_STATIC_PTR(Scm_UVectorClass),
    SCM_CLASS_STATIC_PTR(Scm_SequenceClass),
    SCM_CLASS_STATIC_PTR(Scm_CollectionClass),
    SCM_CLASS_STATIC_PTR(Scm_TopClass),
    NULL
};

SCM_DEFINE_BUILTIN_CLASS(Scm_UVectorClass, NULL, NULL, NULL, NULL,
                         uvector_cpl+1);

static int uvector_index(ScmClass *klass)
{
    if (SCM_EQ(klass, SCM_CLASS_S8VECTOR))  return 0;
    if (SCM_EQ(klass, SCM_CLASS_U8VECTOR))  return 1;
    if (SCM_EQ(klass, SCM_CLASS_S16VECTOR)) return 2;
    if (SCM_EQ(klass, SCM_CLASS_U16VECTOR)) return 3;
    if (SCM_EQ(klass, SCM_CLASS_S32VECTOR)) return 4;
    if (SCM_EQ(klass, SCM_CLASS_U32VECTOR)) return 5;
    if (SCM_EQ(klass, SCM_CLASS_S64VECTOR)) return 6;
    if (SCM_EQ(klass, SCM_CLASS_U64VECTOR)) return 7;
    if (SCM_EQ(klass, SCM_CLASS_F32VECTOR)) return 8;
    if (SCM_EQ(klass, SCM_CLASS_F64VECTOR)) return 9;
    else return -1;
}

/* Returns the size of element of the uvector of given class */
int Scm_UVectorElementSize(ScmClass *klass)
{
    static int sizes[] = { 1, 1, 2, 2, 4, 4, 8, 8,
                           sizeof(float), sizeof(double) };
    int ind = uvector_index(klass);
    if (ind >= 0) return sizes[ind];
    return -1;
}

/* Generic constructor */
ScmObj Scm_MakeUVectorFull(ScmClass *klass, int size, void *init, int immutable, void *owner)
{
    ScmUVector *vec;
    int eltsize = Scm_UVectorElementSize(klass);
    SCM_ASSERT(eltsize >= 1);
    vec = SCM_NEW(ScmUVector);
    SCM_SET_CLASS(vec, klass);
    if (init) {
        vec->elements = init;   /* trust the caller */
    } else {
        vec->elements = SCM_NEW_ATOMIC2(void*, size*eltsize);
    }
    vec->size = size;
    vec->immutable = immutable;
    vec->owner = owner;
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeUVector(ScmClass *klass, int size, void *init)
{
    return Scm_MakeUVectorFull(klass, size, init, FALSE, NULL);
}

/*
 * Generic aliasing
 */
ScmObj Scm_UVectorAlias(ScmClass *klass, ScmUVector *v, int start, int end)
{
    int len = SCM_UVECTOR_SIZE(v), reqalign, srcalign, dstsize;

    SCM_CHECK_START_END(start, end, len);
    reqalign = Scm_UVectorElementSize(klass);
    srcalign = Scm_UVectorElementSize(Scm_ClassOf(SCM_OBJ(v)));
    if (reqalign < 0) {
        Scm_Error("uvector-alias requires uniform vector class, but got %S",
                  klass);
    }
    if ((start*srcalign)%reqalign != 0 || (end*srcalign)%reqalign != 0) {
        Scm_Error("aliasing %S of range (%d, %d) to %S doesn't satisfy alignemnt requirement.",
                  Scm_ClassOf(SCM_OBJ(v)), start, end, klass);
    }
    if (reqalign >= srcalign) dstsize = (end-start) / (reqalign/srcalign);
    else dstsize = (end-start) * (srcalign/reqalign);
    SCM_RETURN(Scm_MakeUVectorFull(klass,
                                   dstsize,
                                   (char*)v->elements + start*srcalign,
                                   SCM_UVECTOR_IMMUTABLE_P(v),
                                   SCM_UVECTOR_OWNER(v)));
}

/*===========================================================
 * Helper functions
 */

static void size_mismatch(const char *name, ScmObj x, ScmObj y)
{
    Scm_Error("%s: argument object sizes do not match: %S vs %S", name, x, y);
}

/* many numeric op procedures takes either uvector, vector, list or
   a constant number as the second arg.  this factors out the common
   code. */
typedef enum {
    ARGTYPE_UVECTOR,
    ARGTYPE_VECTOR,
    ARGTYPE_LIST,
    ARGTYPE_CONST
} ArgType;

static ArgType arg2_check(const char *name, ScmObj x, ScmObj y, int const_ok)
{
    int size = SCM_UVECTOR_SIZE(x);
    if (SCM_UVECTORP(y)) {
        if (SCM_UVECTOR_SIZE(y) != size) size_mismatch(name, SCM_OBJ(x), y);
        return ARGTYPE_UVECTOR;
    } else if (SCM_VECTORP(y)) {
        if (SCM_VECTOR_SIZE(y) != size) size_mismatch(name, SCM_OBJ(x), y);
        return ARGTYPE_VECTOR;
    } else if (SCM_LISTP(y)) {
        if (Scm_Length(y) != size) size_mismatch(name, SCM_OBJ(x), y);
        return ARGTYPE_LIST;
    } else if (!const_ok) {
        Scm_Error("%s: second operand must be either a matching uvector, a vector, or a list, but got %S", y);
    } else if (!SCM_REALP(y)) {
        Scm_Error("%s: second operand must be either a matching uvector, a vector, a list or a number, but got %S", y);
    }
    return ARGTYPE_CONST;
}

/*--------------------------------------------------------
 * Comparison and print
 */
static inline int int64eqv(ScmInt64 x, ScmInt64 y)
{
#if SCM_EMULATE_INT64
    return (x.hi == y.hi && x.lo == y.lo);
#else
    return x == y;
#endif
}

static inline int uint64eqv(ScmUInt64 x, ScmUInt64 y)
{
#if SCM_EMULATE_INT64
    return (x.hi == y.hi && x.lo == y.lo);
#else
    return x == y;
#endif
}

static inline void int64print(ScmPort *out, ScmInt64 v)
{
#if SCM_EMULATE_INT64
    Scm_Printf(out, "%S", Scm_MakeInteger64(v));
#elif SIZEOF_LONG == 4
    char buf[50];
    snprintf(buf, 50, "%lld", v);
    Scm_Printf(out, "%s", buf);
#else
    Scm_Printf(out, "%ld", v);
#endif
}

static inline void uint64print(ScmPort *out, ScmUInt64 v)
{
#if SCM_EMULATE_INT64
    Scm_Printf(out, "%S", Scm_MakeIntegerU64(v));
#elif SIZEOF_LONG == 4
    char buf[50];
    snprintf(buf, 50, "%llu", v);
    Scm_Printf(out, "%s", buf);
#else
    Scm_Printf(out, "%lu", v);
#endif
}

/****** Add, sub, mul and div (f32 and f64 only) *****/
#define s8s8_add(x, y, clamp)   clamp_s8(x+y, clamp)
#define s8s8_sub(x, y, clamp)   clamp_s8(x-y, clamp)
#define s8s8_mul(x, y, clamp)   clamp_s8(x*y, clamp)

static inline long s8g_add(long x, long y, int clamp)
{
    if (y > 255)  return range_s8hi(0, clamp);
    if (y < -256) return range_s8lo(0, clamp);
    return clamp_s8(x+y, clamp);
}

static inline long s8g_sub(long x, long y, int clamp)
{
    if (y < -255) return range_s8hi(0, clamp);
    if (y > 256)  return range_s8lo(0, clamp);
    return clamp_s8(x-y, clamp);
}

static inline long s8g_mul(long x, long y, int clamp)
{
    if (x == 0) return 0;
    if (y > 128) return (x>0)?range_s8hi(0, clamp):range_s8lo(0, clamp);
    return clamp_s8(x*y, clamp);
}

#define u8u8_add(x, y, clamp)   clamp_u8(x+y, clamp)
#define u8u8_sub(x, y, clamp)   clamp_u8((long)(x-y), clamp)
#define u8u8_mul(x, y, clamp)   clamp_u8(x*y, clamp)

static inline u_long u8g_add(u_long x, u_long y, int clamp)
{
    if (y > 255)  return range_u8hi(0, clamp);
    return clamp_u8(x+y, clamp);
}

static inline u_long u8g_sub(u_long x, u_long y, int clamp)
{
    if (y > x)   return range_u8lo(0, clamp);
    return x-y;                 /* never overflows */
}

static inline u_long u8g_mul(u_long x, u_long y, int clamp)
{
    if (x == 0) return 0;
    if (y > 255) return range_u8hi(0, clamp);
    return clamp_u8(x*y, clamp);
}

#define s16s16_add(x, y, clamp)   clamp_s16(x+y, clamp)
#define s16s16_sub(x, y, clamp)   clamp_s16(x-y, clamp)
#define s16s16_mul(x, y, clamp)   clamp_s16(x*y, clamp)

static inline long s16g_add(long x, long y, int clamp)
{
    if (y > 65535)  return range_s16hi(0, clamp);
    if (y < -65536) return range_s16lo(0, clamp);
    return clamp_s16(x+y, clamp);
}

static inline long s16g_sub(long x, long y, int clamp)
{
    if (y < -65535) return range_s16hi(0, clamp);
    if (y > 65536)  return range_s16lo(0, clamp);
    return clamp_s16(x-y, clamp);
}

static inline long s16g_mul(long x, long y, int clamp)
{
    if (x == 0) return 0;
    if (y > 32767) return (x>0)?range_s16hi(0, clamp):range_s16lo(0, clamp);
    return clamp_s16(x*y, clamp);
}

#define u16u16_add(x, y, clamp)   clamp_u16(x+y, clamp)
#define u16u16_sub(x, y, clamp)   clamp_u16((long)(x-y), clamp)
#define u16u16_mul(x, y, clamp)   clamp_u16(x*y, clamp)

static inline u_long u16g_add(u_long x, u_long y, int clamp)
{
    if (y > 65535)  return range_u16hi(0, clamp);
    return clamp_u16(x+y, clamp);
}

static inline u_long u16g_sub(u_long x, u_long y, int clamp)
{
    if (y > x)   return range_u16lo(0, clamp);
    return x-y;                 /* never overflows */
}

static inline u_long u16g_mul(u_long x, u_long y, int clamp)
{
    if (x == 0) return 0;
    if (y > 65535) return range_u16hi(0, clamp);
    return clamp_u16(x*y, clamp);
}

#if SIZEOF_LONG == 4
#define s32s32_add(x, y, clamp)  s32_add_safe(x, y, clamp)
#define s32s32_sub(x, y, clamp)  s32_sub_safe(x, y, clamp)
#define s32s32_mul(x, y, clamp)  s32_mul_safe(x, y, clamp)
#define s32g_add(x, y, clamp)    s32_add_safe(x, y, clamp)
#define s32g_sub(x, y, clamp)    s32_sub_safe(x, y, clamp)
#define s32g_mul(x, y, clamp)    s32_mul_safe(x, y, clamp)
#else  /* SIZEOF_LONG >= 8 */
#define s32s32_add(x, y, clamp)  clamp_s32(x+y, clamp)
#define s32s32_sub(x, y, clamp)  clamp_s32(x-y, clamp)
#define s32s32_mul(x, y, clamp)  clamp_s32(x*y, clamp)
#define s32g_add(x, y, clamp)    s32_add_safe(x, y, clamp)
#define s32g_sub(x, y, clamp)    s32_sub_safe(x, y, clamp)
#define s32g_mul(x, y, clamp)    s32_mul_safe(x, y, clamp)
#endif /* SIZEOF_LONG >= 8 */

static inline long s32_add_safe(long x, long y, int clamp)
{
    long r, v;
    SADDOV(r, v, x, y);
    if (v == 0) return clamp_s32(r, clamp);
    if (v > 0)  return range_s32hi(0, clamp);
    else        return range_s32lo(0, clamp);
}


static inline long s32_sub_safe(long x, long y, int clamp)
{
    long r, v;
    SSUBOV(r, v, x, y);
    if (v == 0) return clamp_s32(r, clamp);
    if (v > 0)  return range_s32hi(0, clamp);
    else        return range_s32lo(0, clamp);
}

static inline long s32_mul_safe(long x, long y, int clamp)
{
    long r, v;
    SMULOV(r, v, x, y);
    if (v == 0) return clamp_s32(r, clamp);
    if (v > 0)  return range_s32hi(0, clamp);
    else        return range_s32lo(0, clamp);
}

#if SIZEOF_LONG == 4
#define u32u32_add(x, y, clamp)  u32_add_safe(x, y, clamp)
#define u32u32_sub(x, y, clamp)  u32_sub_safe(x, y, clamp)
#define u32u32_mul(x, y, clamp)  u32_mul_safe(x, y, clamp)
#define u32g_add(x, y, clamp)    u32_add_safe(x, y, clamp)
#define u32g_sub(x, y, clamp)    u32_sub_safe(x, y, clamp)
#define u32g_mul(x, y, clamp)    u32_mul_safe(x, y, clamp)
#else  /* SIZEOF_LONG >= 8 */
#define u32u32_add(x, y, clamp)  clamp_u32(x+y, clamp)
#define u32u32_sub(x, y, clamp)  u32_sub_safe(x, y, clamp)
#define u32u32_mul(x, y, clamp)  clamp_u32(x*y, clamp)
#define u32g_add(x, y, clamp)    u32_add_safe(x, y, clamp)
#define u32g_sub(x, y, clamp)    u32_sub_safe(x, y, clamp)
#define u32g_mul(x, y, clamp)    u32_mul_safe(x, y, clamp)
#endif /* SIZEOF_LONG >= 8 */

static inline u_long u32_add_safe(u_long x, u_long y, int clamp)
{
    u_long r, v;
    UADDOV(r, v, x, y);
    if (v == 0) return clamp_u32(r, clamp);
    else        return range_u32hi(0, clamp);
}

static inline u_long u32_sub_safe(u_long x, u_long y, int clamp)
{
    u_long r, v;
    USUBOV(r, v, x, y);
    if (v == 0) return clamp_u32(r, clamp);
    else        return range_u32lo(0, clamp);
}

static inline u_long u32_mul_safe(u_long x, u_long y, int clamp)
{
    u_long r, v;
    UMULOV(r, v, x, y);
    if (v == 0) return clamp_u32(r, clamp);
    else        return range_u32hi(0, clamp);
}

#define s64s64_add(x, y, clamp)  s64g_add(x, y, clamp)
#define s64s64_sub(x, y, clamp)  s64g_sub(x, y, clamp)
#define s64s64_mul(x, y, clamp)  s64g_mul(x, y, clamp)

static inline ScmInt64 s64g_add(ScmInt64 x, ScmInt64 y, int clamp)
{
#if SIZEOF_LONG == 4
    ScmObj xx = Scm_MakeInteger64(x);
    ScmObj yy = Scm_MakeInteger64(y);
    ScmObj r = Scm_Add(xx, yy);
    return Scm_GetInteger64Clamp(r, clamp, NULL);
#else
    long r, v;
    SADDOV(r, v, x, y);
    if (v == 0) return r;
    if (v > 0)  return range_s64hi(0, clamp);
    else        return range_s64lo(0, clamp);
#endif
}

static inline ScmInt64 s64g_sub(ScmInt64 x, ScmInt64 y, int clamp)
{
#if SIZEOF_LONG == 4
    ScmObj xx = Scm_MakeInteger64(x);
    ScmObj yy = Scm_MakeInteger64(y);
    ScmObj r = Scm_Sub(xx, yy);
    return Scm_GetInteger64Clamp(r, clamp, NULL);
#else
    long r, v;
    SSUBOV(r, v, x, y);
    if (v == 0) return r;
    if (v > 0)  return range_s64hi(0, clamp);
    else        return range_s64lo(0, clamp);
#endif
}

static inline ScmInt64 s64g_mul(ScmInt64 x, ScmInt64 y, int clamp)
{
#if SIZEOF_LONG == 4
    ScmObj xx = Scm_MakeInteger64(x);
    ScmObj yy = Scm_MakeInteger64(y);
    ScmObj r = Scm_Mul(xx, yy);
    return Scm_GetInteger64Clamp(r, clamp, NULL);
#else
    long r, v;
    SMULOV(r, v, x, y);
    if (v == 0) return r;
    if (v > 0)  return range_s64hi(0, clamp);
    else        return range_s64lo(0, clamp);
#endif
}

#define u64u64_add(x, y, clamp)  u64g_add(x, y, clamp)
#define u64u64_sub(x, y, clamp)  u64g_sub(x, y, clamp)
#define u64u64_mul(x, y, clamp)  u64g_mul(x, y, clamp)

static inline ScmUInt64 u64g_add(ScmUInt64 x, ScmUInt64 y, int clamp)
{
#if SIZEOF_LONG == 4
    ScmObj xx = Scm_MakeIntegerU64(x);
    ScmObj yy = Scm_MakeIntegerU64(y);
    ScmObj r = Scm_Add(xx, yy);
    return Scm_GetIntegerU64Clamp(r, clamp, NULL);
#else
    u_long r, v;
    UADDOV(r, v, x, y);
    if (v == 0) return r;
    else        return range_u64hi(0, clamp);
#endif
}

static inline ScmUInt64 u64g_sub(ScmUInt64 x, ScmUInt64 y, int clamp)
{
#if SIZEOF_LONG == 4
    ScmObj xx = Scm_MakeIntegerU64(x);
    ScmObj yy = Scm_MakeIntegerU64(y);
    ScmObj r = Scm_Sub(xx, yy);
    return Scm_GetIntegerU64Clamp(r, clamp, NULL);
#else
    u_long r, v;
    USUBOV(r, v, x, y);
    if (v == 0) return r;
    else        return range_u64lo(0, clamp);
#endif
}

static inline ScmUInt64 u64g_mul(ScmUInt64 x, ScmUInt64 y, int clamp)
{
#if SIZEOF_LONG == 4
    ScmObj xx = Scm_MakeIntegerU64(x);
    ScmObj yy = Scm_MakeIntegerU64(y);
    ScmObj r = Scm_Mul(xx, yy);
    return Scm_GetIntegerU64Clamp(r, clamp, NULL);
#else
    u_long r, v;
    UMULOV(r, v, x, y);
    if (v == 0) return r;
    else        return range_u64hi(0, clamp);
#endif
}

#define f32f32_add(x, y, clamp)   (x+y)
#define f32f32_sub(x, y, clamp)   (x-y)
#define f32f32_mul(x, y, clamp)   (x*y)
#define f32f32_div(x, y, clamp)   (x/y)

#define f32g_add(x, y, clamp)   (x+y)
#define f32g_sub(x, y, clamp)   (x-y)
#define f32g_mul(x, y, clamp)   (x*y)
#define f32g_div(x, y, clamp)   (x/y)

#define f64f64_add(x, y, clamp)   (x+y)
#define f64f64_sub(x, y, clamp)   (x-y)
#define f64f64_mul(x, y, clamp)   (x*y)
#define f64f64_div(x, y, clamp)   (x/y)

#define f64g_add(x, y, clamp)   (x+y)
#define f64g_sub(x, y, clamp)   (x-y)
#define f64g_mul(x, y, clamp)   (x*y)
#define f64g_div(x, y, clamp)   (x/y)

/****** Number extraction *****/
/* like unbox, but not as strict.  sets *oor = TRUE if x is out of range. */

static inline long s8num(ScmObj x, int *oor)
{
    return Scm_GetIntegerClamp(x, SCM_CLAMP_NONE, oor);
}

#define s16num(x, oor)           s8num(x, oor)
#define s32num(x, oor)           s8num(x, oor)

static inline u_long u8num(ScmObj x, int *oor)
{
    return Scm_GetIntegerUClamp(x, SCM_CLAMP_NONE, oor);
}

#define u16num(x, oor)           u8num(x, oor)
#define u32num(x, oor)           u8num(x, oor)

static inline ScmInt64 s64num(ScmObj x, int *oor)
{
    return Scm_GetInteger64Clamp(x, SCM_CLAMP_NONE, oor);
}

static inline ScmUInt64 u64num(ScmObj x, int *oor)
{
    return Scm_GetIntegerU64Clamp(x, SCM_CLAMP_NONE, oor);
}

#define f32num(x, oor)           ((*oor = FALSE),((float)Scm_GetDouble(x)))
#define f64num(x, oor)           ((*oor = FALSE), Scm_GetDouble(x))
/****** Bit operation *****/

static inline u_long bitext(ScmObj x)
{
    if (SCM_INTP(x)) return (u_long)SCM_INT_VALUE(x);
    if (SCM_BIGNUMP(x)) {
        if (SCM_BIGNUM_SIGN(x) > 0) {
            return SCM_BIGNUM(x)->values[0];
        } else {
            return ~(SCM_BIGNUM(x)->values[0]) + 1;
        }
    }
    Scm_Error("integer required, but got %S", x);
    return 0;
}

static inline ScmUInt64 bitext64(ScmObj x)
{
#if SCM_EMULATE_INT64
    ScmUInt64 r = {0, 0};
    if (SCM_INTP(x)) r.lo = SCM_INT_VALUE(x);
    else if (SCM_BIGNUMP(x)) {
        ScmObj xx = Scm_LogAnd(x, SCM_2_64_MINUS_1);
        ScmUInt64 ii = Scm_GetIntegerU64(xx);
        r.lo = ii.lo;
        r.hi = ii.hi;
    }
    else goto type_err;
#else
    ScmUInt64 r = 0;
    if (SCM_INTP(x)) r = SCM_INT_VALUE(x);
    else if (SCM_BIGNUMP(x)) {
        ScmObj xx = Scm_LogAnd(x, SCM_2_64_MINUS_1);
        r = Scm_GetIntegerU64(xx);
    }
    else goto type_err;
#endif
    return r;
  type_err:
    Scm_Error("integer required, but got %S", x);
    return r;                   /* dummy */
}

#if SCM_EMULATE_INT64
#define INT64BITOP(r, x, op, y)  ((r.lo = x.lo op y.lo), (r.hi = x.hi op y.hi))
#else
#define INT64BITOP(r, x, op, y)  (r = x op y)
#endif
/****** Multiply-and-add operation. *****/

static inline long s8muladd(long x, long y, long acc, ScmObj *sacc)
{
    long k, v, m;
    m = x * y;
    SADDOV(k, v, acc, m);
    if (v) {
        *sacc = Scm_Add(*sacc, Scm_MakeInteger(acc));
        return m;
    } else {
        return k;
    }
}

#define s16muladd(x, y, acc, sacc) s8muladd(x, y, acc, sacc)

static inline long s32muladd(long x, long y, long acc, ScmObj *sacc)
{
    long k, v, m;
    SMULOV(m, v, x, y);
    if (v) {
        *sacc = Scm_Add(*sacc, Scm_Mul(Scm_MakeInteger(x),
                                       Scm_MakeInteger(y)));
        return acc;
    } else {
        SADDOV(k, v, acc, m);
        if (v) {
            *sacc = Scm_Add(*sacc, Scm_MakeInteger(acc));
            return m;
        } else {
            return k;
        }
    }
}

#if SIZEOF_LONG == 4
static inline ScmInt64 s64muladd(ScmInt64 x, ScmInt64 y, ScmInt64 acc, ScmObj *sacc)
{
    /* we don't use acc, and operate only on sacc. */
    *sacc = Scm_Add(*sacc, Scm_Mul(Scm_MakeInteger64(x),
                                   Scm_MakeInteger64(y)));
    return acc;
}
#else
#define s64muladd(x, y, acc, sacc) s32muladd(x, y, acc, sacc)
#endif

static inline u_long u8muladd(u_long x, u_long y, u_long acc, ScmObj *sacc)
{
    u_long k, v, m;
    m = x * y;
    UADDOV(k, v, acc, m);
    if (v) {
        *sacc = Scm_Add(*sacc, Scm_MakeIntegerU(acc));
        return m;
    } else {
        return k;
    }
}

#define u16muladd(x, y, acc, sacc) u8muladd(x, y, acc, sacc)

static inline u_long u32muladd(u_long x, u_long y, u_long acc, ScmObj *sacc)
{
    u_long k, v, m;
    UMULOV(m, v, x, y);
    if (v) {
        *sacc = Scm_Add(*sacc, Scm_Mul(Scm_MakeIntegerU(x),
                                              Scm_MakeIntegerU(y)));
        return acc;
    } else {
        UADDOV(k, v, acc, m);
        if (v) {
            *sacc = Scm_Add(*sacc, Scm_MakeIntegerU(acc));
            return m;
        } else {
            return k;
        }
    }
}

#if SIZEOF_LONG == 4
static inline ScmUInt64 u64muladd(ScmUInt64 x, ScmUInt64 y, ScmUInt64 acc, ScmObj *sacc)
{
    /* we don't use acc, and operate only on sacc. */
    *sacc = Scm_Add(*sacc, Scm_Mul(Scm_MakeIntegerU64(x),
                                          Scm_MakeIntegerU64(y)));
    return acc;
}
#else
#define u64muladd(x, y, acc, sacc) u32muladd(x, y, acc, sacc)
#endif

#define f32muladd(x, y, acc, sacc)  (acc + x*y)
#define f64muladd(x, y, acc, sacc)  (acc + x*y)

#if SCM_EMULATE_INT64
#define INT64LT(a, b)  ((a.hi < b.hi) || (a.hi == b.hi) && (a.lo < b.lo))
#else
#define INT64LT(a, b)  (a < b)
#endif


#define SWAP(x, y)   (t = dd.c[x], dd.c[x] = dd.c[y], dd.c[y] = t)

static void swapb16(unsigned short *loc)
{
    union {
        unsigned short s;
        unsigned char c[2];
    } dd;
    unsigned char t;
    dd.s = *loc;
    SWAP(0, 1);
    *loc = dd.s;
}

static void swapb32(ScmUInt32 *loc)
{
    union {
        ScmUInt32 i;
        unsigned char c[4];
    } dd;
    unsigned char t;
    dd.i = *loc;
    SWAP(0, 3);
    SWAP(1, 2);
    *loc = dd.i;
}

static void swapb64(ScmUInt64 *loc)
{
    union {
        ScmUInt64 l;
        unsigned char c[4];
    } dd;
    unsigned char t;
    dd.l = *loc;
    SWAP(0, 7);
    SWAP(1, 6);
    SWAP(2, 5);
    SWAP(3, 4);
    *loc = dd.l;
}


/*---------------------------------------------------------------
 * S8Vector
 */

/*
 * Class stuff
 */

static void print_s8vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#s8(");
    for (i=0; i<SCM_S8VECTOR_SIZE(obj); i++) {
        signed char elt = SCM_S8VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%d", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_s8vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_S8VECTOR_SIZE(x), i;
    signed char xx, yy;
    if (SCM_S8VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_S8VECTOR_ELEMENTS(x)[i];
        yy = SCM_S8VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_S8VectorClass,
                         print_s8vector, compare_s8vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmS8Vector *make_s8vector(int size, signed char *eltp)
{
    return (ScmS8Vector*)Scm_MakeUVector(SCM_CLASS_S8VECTOR, size, eltp);
}

ScmObj Scm_MakeS8Vector(int size, signed char fill)
{
    ScmS8Vector *vec = make_s8vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS8VectorFromArray(int size, const signed char array[])
{
    ScmS8Vector *vec = make_s8vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS8VectorFromArrayShared(int size, signed char array[])
{
    ScmS8Vector *vec = make_s8vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToS8Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmS8Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_s8vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        signed char elt;
        ScmObj obj = SCM_CAR(cp);
        elt = s8unbox(obj, clamp);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToS8Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmS8Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_s8vector(end-start, NULL);
    for (i=start; i<end; i++) {
        signed char elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = s8unbox(obj, clamp);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_S8VectorFill(ScmS8Vector *vec, signed char fill, int start, int end)
{
    int i, size = SCM_S8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_S8VectorRef(ScmS8Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    signed char elt;
    if (index < 0 || index >= SCM_S8VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = SCM_MAKE_INT(elt);
    return r;
}

ScmObj Scm_S8VectorSet(ScmS8Vector *vec, int index, ScmObj val, int clamp)
{
    signed char elt;
    if (index < 0 || index >= SCM_S8VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = s8unbox(val, clamp);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_S8VectorToList(ScmS8Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_S8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        signed char elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_S8VectorToVector(ScmS8Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_S8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        signed char elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_S8VectorCopy(ScmS8Vector *vec, int start, int end)
{
    int size = SCM_S8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeS8VectorFromArray(end-start,
                                     SCM_S8VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_S8VectorCopyX(ScmS8Vector *dst,
                           int dstart,
                           ScmS8Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_S8VECTOR_SIZE(dst);
    int slen = SCM_S8VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_S8VECTOR_ELEMENTS(dst) + dstart,
           SCM_S8VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(signed char));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * U8Vector
 */

/*
 * Class stuff
 */

static void print_u8vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#u8(");
    for (i=0; i<SCM_U8VECTOR_SIZE(obj); i++) {
        unsigned char elt = SCM_U8VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%d", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_u8vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_U8VECTOR_SIZE(x), i;
    unsigned char xx, yy;
    if (SCM_U8VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_U8VECTOR_ELEMENTS(x)[i];
        yy = SCM_U8VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_U8VectorClass,
                         print_u8vector, compare_u8vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmU8Vector *make_u8vector(int size, unsigned char *eltp)
{
    return (ScmU8Vector*)Scm_MakeUVector(SCM_CLASS_U8VECTOR, size, eltp);
}

ScmObj Scm_MakeU8Vector(int size, unsigned char fill)
{
    ScmU8Vector *vec = make_u8vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU8VectorFromArray(int size, const unsigned char array[])
{
    ScmU8Vector *vec = make_u8vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU8VectorFromArrayShared(int size, unsigned char array[])
{
    ScmU8Vector *vec = make_u8vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToU8Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmU8Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_u8vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        unsigned char elt;
        ScmObj obj = SCM_CAR(cp);
        elt = u8unbox(obj, clamp);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToU8Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmU8Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_u8vector(end-start, NULL);
    for (i=start; i<end; i++) {
        unsigned char elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = u8unbox(obj, clamp);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_U8VectorFill(ScmU8Vector *vec, unsigned char fill, int start, int end)
{
    int i, size = SCM_U8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_U8VectorRef(ScmU8Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    unsigned char elt;
    if (index < 0 || index >= SCM_U8VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = SCM_MAKE_INT(elt);
    return r;
}

ScmObj Scm_U8VectorSet(ScmU8Vector *vec, int index, ScmObj val, int clamp)
{
    unsigned char elt;
    if (index < 0 || index >= SCM_U8VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = u8unbox(val, clamp);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_U8VectorToList(ScmU8Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_U8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        unsigned char elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_U8VectorToVector(ScmU8Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_U8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        unsigned char elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_U8VectorCopy(ScmU8Vector *vec, int start, int end)
{
    int size = SCM_U8VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeU8VectorFromArray(end-start,
                                     SCM_U8VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_U8VectorCopyX(ScmU8Vector *dst,
                           int dstart,
                           ScmU8Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_U8VECTOR_SIZE(dst);
    int slen = SCM_U8VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_U8VECTOR_ELEMENTS(dst) + dstart,
           SCM_U8VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(unsigned char));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * S16Vector
 */

/*
 * Class stuff
 */

static void print_s16vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#s16(");
    for (i=0; i<SCM_S16VECTOR_SIZE(obj); i++) {
        short elt = SCM_S16VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%d", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_s16vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_S16VECTOR_SIZE(x), i;
    short xx, yy;
    if (SCM_S16VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_S16VECTOR_ELEMENTS(x)[i];
        yy = SCM_S16VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_S16VectorClass,
                         print_s16vector, compare_s16vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmS16Vector *make_s16vector(int size, short *eltp)
{
    return (ScmS16Vector*)Scm_MakeUVector(SCM_CLASS_S16VECTOR, size, eltp);
}

ScmObj Scm_MakeS16Vector(int size, short fill)
{
    ScmS16Vector *vec = make_s16vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS16VectorFromArray(int size, const short array[])
{
    ScmS16Vector *vec = make_s16vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS16VectorFromArrayShared(int size, short array[])
{
    ScmS16Vector *vec = make_s16vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToS16Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmS16Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_s16vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        short elt;
        ScmObj obj = SCM_CAR(cp);
        elt = s16unbox(obj, clamp);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToS16Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmS16Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_s16vector(end-start, NULL);
    for (i=start; i<end; i++) {
        short elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = s16unbox(obj, clamp);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_S16VectorFill(ScmS16Vector *vec, short fill, int start, int end)
{
    int i, size = SCM_S16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_S16VectorRef(ScmS16Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    short elt;
    if (index < 0 || index >= SCM_S16VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = SCM_MAKE_INT(elt);
    return r;
}

ScmObj Scm_S16VectorSet(ScmS16Vector *vec, int index, ScmObj val, int clamp)
{
    short elt;
    if (index < 0 || index >= SCM_S16VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = s16unbox(val, clamp);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_S16VectorToList(ScmS16Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_S16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        short elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_S16VectorToVector(ScmS16Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_S16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        short elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_S16VectorCopy(ScmS16Vector *vec, int start, int end)
{
    int size = SCM_S16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeS16VectorFromArray(end-start,
                                     SCM_S16VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_S16VectorCopyX(ScmS16Vector *dst,
                           int dstart,
                           ScmS16Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_S16VECTOR_SIZE(dst);
    int slen = SCM_S16VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_S16VECTOR_ELEMENTS(dst) + dstart,
           SCM_S16VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(short));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * U16Vector
 */

/*
 * Class stuff
 */

static void print_u16vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#u16(");
    for (i=0; i<SCM_U16VECTOR_SIZE(obj); i++) {
        unsigned short elt = SCM_U16VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%d", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_u16vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_U16VECTOR_SIZE(x), i;
    unsigned short xx, yy;
    if (SCM_U16VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_U16VECTOR_ELEMENTS(x)[i];
        yy = SCM_U16VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_U16VectorClass,
                         print_u16vector, compare_u16vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmU16Vector *make_u16vector(int size, unsigned short *eltp)
{
    return (ScmU16Vector*)Scm_MakeUVector(SCM_CLASS_U16VECTOR, size, eltp);
}

ScmObj Scm_MakeU16Vector(int size, unsigned short fill)
{
    ScmU16Vector *vec = make_u16vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU16VectorFromArray(int size, const unsigned short array[])
{
    ScmU16Vector *vec = make_u16vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU16VectorFromArrayShared(int size, unsigned short array[])
{
    ScmU16Vector *vec = make_u16vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToU16Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmU16Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_u16vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        unsigned short elt;
        ScmObj obj = SCM_CAR(cp);
        elt = u16unbox(obj, clamp);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToU16Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmU16Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_u16vector(end-start, NULL);
    for (i=start; i<end; i++) {
        unsigned short elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = u16unbox(obj, clamp);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_U16VectorFill(ScmU16Vector *vec, unsigned short fill, int start, int end)
{
    int i, size = SCM_U16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_U16VectorRef(ScmU16Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    unsigned short elt;
    if (index < 0 || index >= SCM_U16VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = SCM_MAKE_INT(elt);
    return r;
}

ScmObj Scm_U16VectorSet(ScmU16Vector *vec, int index, ScmObj val, int clamp)
{
    unsigned short elt;
    if (index < 0 || index >= SCM_U16VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = u16unbox(val, clamp);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_U16VectorToList(ScmU16Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_U16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        unsigned short elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_U16VectorToVector(ScmU16Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_U16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        unsigned short elt = vec->elements[i];
        obj = SCM_MAKE_INT(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_U16VectorCopy(ScmU16Vector *vec, int start, int end)
{
    int size = SCM_U16VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeU16VectorFromArray(end-start,
                                     SCM_U16VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_U16VectorCopyX(ScmU16Vector *dst,
                           int dstart,
                           ScmU16Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_U16VECTOR_SIZE(dst);
    int slen = SCM_U16VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_U16VECTOR_ELEMENTS(dst) + dstart,
           SCM_U16VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(unsigned short));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * S32Vector
 */

/*
 * Class stuff
 */

static void print_s32vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#s32(");
    for (i=0; i<SCM_S32VECTOR_SIZE(obj); i++) {
        ScmInt32 elt = SCM_S32VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%d", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_s32vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_S32VECTOR_SIZE(x), i;
    ScmInt32 xx, yy;
    if (SCM_S32VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_S32VECTOR_ELEMENTS(x)[i];
        yy = SCM_S32VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_S32VectorClass,
                         print_s32vector, compare_s32vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmS32Vector *make_s32vector(int size, ScmInt32 *eltp)
{
    return (ScmS32Vector*)Scm_MakeUVector(SCM_CLASS_S32VECTOR, size, eltp);
}

ScmObj Scm_MakeS32Vector(int size, ScmInt32 fill)
{
    ScmS32Vector *vec = make_s32vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS32VectorFromArray(int size, const ScmInt32 array[])
{
    ScmS32Vector *vec = make_s32vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS32VectorFromArrayShared(int size, ScmInt32 array[])
{
    ScmS32Vector *vec = make_s32vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToS32Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmS32Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_s32vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        ScmInt32 elt;
        ScmObj obj = SCM_CAR(cp);
        elt = Scm_GetInteger32Clamp(obj, clamp, NULL);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToS32Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmS32Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_s32vector(end-start, NULL);
    for (i=start; i<end; i++) {
        ScmInt32 elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = Scm_GetInteger32Clamp(obj, clamp, NULL);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_S32VectorFill(ScmS32Vector *vec, ScmInt32 fill, int start, int end)
{
    int i, size = SCM_S32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_S32VectorRef(ScmS32Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    ScmInt32 elt;
    if (index < 0 || index >= SCM_S32VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = Scm_MakeInteger(elt);
    return r;
}

ScmObj Scm_S32VectorSet(ScmS32Vector *vec, int index, ScmObj val, int clamp)
{
    ScmInt32 elt;
    if (index < 0 || index >= SCM_S32VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = Scm_GetInteger32Clamp(val, clamp, NULL);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_S32VectorToList(ScmS32Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_S32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmInt32 elt = vec->elements[i];
        obj = Scm_MakeInteger(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_S32VectorToVector(ScmS32Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_S32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmInt32 elt = vec->elements[i];
        obj = Scm_MakeInteger(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_S32VectorCopy(ScmS32Vector *vec, int start, int end)
{
    int size = SCM_S32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeS32VectorFromArray(end-start,
                                     SCM_S32VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_S32VectorCopyX(ScmS32Vector *dst,
                           int dstart,
                           ScmS32Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_S32VECTOR_SIZE(dst);
    int slen = SCM_S32VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_S32VECTOR_ELEMENTS(dst) + dstart,
           SCM_S32VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(ScmInt32));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * U32Vector
 */

/*
 * Class stuff
 */

static void print_u32vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#u32(");
    for (i=0; i<SCM_U32VECTOR_SIZE(obj); i++) {
        ScmUInt32 elt = SCM_U32VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%u", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_u32vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_U32VECTOR_SIZE(x), i;
    ScmUInt32 xx, yy;
    if (SCM_U32VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_U32VECTOR_ELEMENTS(x)[i];
        yy = SCM_U32VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_U32VectorClass,
                         print_u32vector, compare_u32vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmU32Vector *make_u32vector(int size, ScmUInt32 *eltp)
{
    return (ScmU32Vector*)Scm_MakeUVector(SCM_CLASS_U32VECTOR, size, eltp);
}

ScmObj Scm_MakeU32Vector(int size, ScmUInt32 fill)
{
    ScmU32Vector *vec = make_u32vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU32VectorFromArray(int size, const ScmUInt32 array[])
{
    ScmU32Vector *vec = make_u32vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU32VectorFromArrayShared(int size, ScmUInt32 array[])
{
    ScmU32Vector *vec = make_u32vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToU32Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmU32Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_u32vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        ScmUInt32 elt;
        ScmObj obj = SCM_CAR(cp);
        elt = Scm_GetIntegerU32Clamp(obj, clamp, NULL);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToU32Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmU32Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_u32vector(end-start, NULL);
    for (i=start; i<end; i++) {
        ScmUInt32 elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = Scm_GetIntegerU32Clamp(obj, clamp, NULL);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_U32VectorFill(ScmU32Vector *vec, ScmUInt32 fill, int start, int end)
{
    int i, size = SCM_U32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_U32VectorRef(ScmU32Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    ScmUInt32 elt;
    if (index < 0 || index >= SCM_U32VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = Scm_MakeIntegerU(elt);
    return r;
}

ScmObj Scm_U32VectorSet(ScmU32Vector *vec, int index, ScmObj val, int clamp)
{
    ScmUInt32 elt;
    if (index < 0 || index >= SCM_U32VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = Scm_GetIntegerU32Clamp(val, clamp, NULL);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_U32VectorToList(ScmU32Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_U32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmUInt32 elt = vec->elements[i];
        obj = Scm_MakeIntegerU(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_U32VectorToVector(ScmU32Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_U32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmUInt32 elt = vec->elements[i];
        obj = Scm_MakeIntegerU(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_U32VectorCopy(ScmU32Vector *vec, int start, int end)
{
    int size = SCM_U32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeU32VectorFromArray(end-start,
                                     SCM_U32VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_U32VectorCopyX(ScmU32Vector *dst,
                           int dstart,
                           ScmU32Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_U32VECTOR_SIZE(dst);
    int slen = SCM_U32VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_U32VECTOR_ELEMENTS(dst) + dstart,
           SCM_U32VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(ScmUInt32));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * S64Vector
 */

/*
 * Class stuff
 */

static void print_s64vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#s64(");
    for (i=0; i<SCM_S64VECTOR_SIZE(obj); i++) {
        ScmInt64 elt = SCM_S64VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        int64print(out, elt);
    }
    Scm_Printf(out, ")");
}

static int compare_s64vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_S64VECTOR_SIZE(x), i;
    ScmInt64 xx, yy;
    if (SCM_S64VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_S64VECTOR_ELEMENTS(x)[i];
        yy = SCM_S64VECTOR_ELEMENTS(y)[i];
        if (!int64eqv(xx, yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_S64VectorClass,
                         print_s64vector, compare_s64vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmS64Vector *make_s64vector(int size, ScmInt64 *eltp)
{
    return (ScmS64Vector*)Scm_MakeUVector(SCM_CLASS_S64VECTOR, size, eltp);
}

ScmObj Scm_MakeS64Vector(int size, ScmInt64 fill)
{
    ScmS64Vector *vec = make_s64vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS64VectorFromArray(int size, const ScmInt64 array[])
{
    ScmS64Vector *vec = make_s64vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeS64VectorFromArrayShared(int size, ScmInt64 array[])
{
    ScmS64Vector *vec = make_s64vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToS64Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmS64Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_s64vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        ScmInt64 elt;
        ScmObj obj = SCM_CAR(cp);
        elt = Scm_GetInteger64Clamp(obj, clamp, NULL);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToS64Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmS64Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_s64vector(end-start, NULL);
    for (i=start; i<end; i++) {
        ScmInt64 elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = Scm_GetInteger64Clamp(obj, clamp, NULL);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_S64VectorFill(ScmS64Vector *vec, ScmInt64 fill, int start, int end)
{
    int i, size = SCM_S64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_S64VectorRef(ScmS64Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    ScmInt64 elt;
    if (index < 0 || index >= SCM_S64VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = Scm_MakeInteger64(elt);
    return r;
}

ScmObj Scm_S64VectorSet(ScmS64Vector *vec, int index, ScmObj val, int clamp)
{
    ScmInt64 elt;
    if (index < 0 || index >= SCM_S64VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = Scm_GetInteger64Clamp(val, clamp, NULL);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_S64VectorToList(ScmS64Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_S64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmInt64 elt = vec->elements[i];
        obj = Scm_MakeInteger64(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_S64VectorToVector(ScmS64Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_S64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmInt64 elt = vec->elements[i];
        obj = Scm_MakeInteger64(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_S64VectorCopy(ScmS64Vector *vec, int start, int end)
{
    int size = SCM_S64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeS64VectorFromArray(end-start,
                                     SCM_S64VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_S64VectorCopyX(ScmS64Vector *dst,
                           int dstart,
                           ScmS64Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_S64VECTOR_SIZE(dst);
    int slen = SCM_S64VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_S64VECTOR_ELEMENTS(dst) + dstart,
           SCM_S64VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(ScmInt64));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * U64Vector
 */

/*
 * Class stuff
 */

static void print_u64vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#u64(");
    for (i=0; i<SCM_U64VECTOR_SIZE(obj); i++) {
        ScmUInt64 elt = SCM_U64VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        uint64print(out, elt);
    }
    Scm_Printf(out, ")");
}

static int compare_u64vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_U64VECTOR_SIZE(x), i;
    ScmUInt64 xx, yy;
    if (SCM_U64VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_U64VECTOR_ELEMENTS(x)[i];
        yy = SCM_U64VECTOR_ELEMENTS(y)[i];
        if (!uint64eqv(xx, yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_U64VectorClass,
                         print_u64vector, compare_u64vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmU64Vector *make_u64vector(int size, ScmUInt64 *eltp)
{
    return (ScmU64Vector*)Scm_MakeUVector(SCM_CLASS_U64VECTOR, size, eltp);
}

ScmObj Scm_MakeU64Vector(int size, ScmUInt64 fill)
{
    ScmU64Vector *vec = make_u64vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU64VectorFromArray(int size, const ScmUInt64 array[])
{
    ScmU64Vector *vec = make_u64vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeU64VectorFromArrayShared(int size, ScmUInt64 array[])
{
    ScmU64Vector *vec = make_u64vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToU64Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmU64Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_u64vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        ScmUInt64 elt;
        ScmObj obj = SCM_CAR(cp);
        elt = Scm_GetIntegerU64Clamp(obj, clamp, NULL);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToU64Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmU64Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_u64vector(end-start, NULL);
    for (i=start; i<end; i++) {
        ScmUInt64 elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = Scm_GetIntegerU64Clamp(obj, clamp, NULL);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_U64VectorFill(ScmU64Vector *vec, ScmUInt64 fill, int start, int end)
{
    int i, size = SCM_U64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_U64VectorRef(ScmU64Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    ScmUInt64 elt;
    if (index < 0 || index >= SCM_U64VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = Scm_MakeIntegerU64(elt);
    return r;
}

ScmObj Scm_U64VectorSet(ScmU64Vector *vec, int index, ScmObj val, int clamp)
{
    ScmUInt64 elt;
    if (index < 0 || index >= SCM_U64VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = Scm_GetIntegerU64Clamp(val, clamp, NULL);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_U64VectorToList(ScmU64Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_U64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmUInt64 elt = vec->elements[i];
        obj = Scm_MakeIntegerU64(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_U64VectorToVector(ScmU64Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_U64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        ScmUInt64 elt = vec->elements[i];
        obj = Scm_MakeIntegerU64(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_U64VectorCopy(ScmU64Vector *vec, int start, int end)
{
    int size = SCM_U64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeU64VectorFromArray(end-start,
                                     SCM_U64VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_U64VectorCopyX(ScmU64Vector *dst,
                           int dstart,
                           ScmU64Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_U64VECTOR_SIZE(dst);
    int slen = SCM_U64VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_U64VECTOR_ELEMENTS(dst) + dstart,
           SCM_U64VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(ScmUInt64));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * F32Vector
 */

/*
 * Class stuff
 */

static void print_f32vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#f32(");
    for (i=0; i<SCM_F32VECTOR_SIZE(obj); i++) {
        float elt = SCM_F32VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_Printf(out, "%.9g", elt);
    }
    Scm_Printf(out, ")");
}

static int compare_f32vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_F32VECTOR_SIZE(x), i;
    float xx, yy;
    if (SCM_F32VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_F32VECTOR_ELEMENTS(x)[i];
        yy = SCM_F32VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_F32VectorClass,
                         print_f32vector, compare_f32vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmF32Vector *make_f32vector(int size, float *eltp)
{
    return (ScmF32Vector*)Scm_MakeUVector(SCM_CLASS_F32VECTOR, size, eltp);
}

ScmObj Scm_MakeF32Vector(int size, float fill)
{
    ScmF32Vector *vec = make_f32vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeF32VectorFromArray(int size, const float array[])
{
    ScmF32Vector *vec = make_f32vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeF32VectorFromArrayShared(int size, float array[])
{
    ScmF32Vector *vec = make_f32vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToF32Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmF32Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_f32vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        float elt;
        ScmObj obj = SCM_CAR(cp);
        elt = (float)Scm_GetDouble(obj);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToF32Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmF32Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_f32vector(end-start, NULL);
    for (i=start; i<end; i++) {
        float elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = (float)Scm_GetDouble(obj);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_F32VectorFill(ScmF32Vector *vec, float fill, int start, int end)
{
    int i, size = SCM_F32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_F32VectorRef(ScmF32Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    float elt;
    if (index < 0 || index >= SCM_F32VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = Scm_MakeFlonum((double)elt);
    return r;
}

ScmObj Scm_F32VectorSet(ScmF32Vector *vec, int index, ScmObj val, int clamp)
{
    float elt;
    if (index < 0 || index >= SCM_F32VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = (float)Scm_GetDouble(val);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_F32VectorToList(ScmF32Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_F32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        float elt = vec->elements[i];
        obj = Scm_MakeFlonum((double)elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_F32VectorToVector(ScmF32Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_F32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        float elt = vec->elements[i];
        obj = Scm_MakeFlonum((double)elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_F32VectorCopy(ScmF32Vector *vec, int start, int end)
{
    int size = SCM_F32VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeF32VectorFromArray(end-start,
                                     SCM_F32VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_F32VectorCopyX(ScmF32Vector *dst,
                           int dstart,
                           ScmF32Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_F32VECTOR_SIZE(dst);
    int slen = SCM_F32VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_F32VECTOR_ELEMENTS(dst) + dstart,
           SCM_F32VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(float));
    return SCM_OBJ(dst);
}


/*---------------------------------------------------------------
 * F64Vector
 */

/*
 * Class stuff
 */

static void print_f64vector(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
{
    int i;
    Scm_Printf(out, "#f64(");
    for (i=0; i<SCM_F64VECTOR_SIZE(obj); i++) {
        double elt = SCM_F64VECTOR_ELEMENTS(obj)[i];
        if (i != 0) Scm_Printf(out, " ");
        Scm_PrintDouble(out, (double)elt, 0);
    }
    Scm_Printf(out, ")");
}

static int compare_f64vector(ScmObj x, ScmObj y, int equalp)
{
    int len = SCM_F64VECTOR_SIZE(x), i;
    double xx, yy;
    if (SCM_F64VECTOR_SIZE(y) != len) return -1;
    for (i=0; i<len; i++) {
        xx = SCM_F64VECTOR_ELEMENTS(x)[i];
        yy = SCM_F64VECTOR_ELEMENTS(y)[i];
        if (!(xx == yy)) {
            return -1;
        }
    }
    return 0;
}

SCM_DEFINE_BUILTIN_CLASS(Scm_F64VectorClass,
                         print_f64vector, compare_f64vector, NULL, NULL,
                         uvector_cpl);

/*
 * Constructor
 */
static ScmF64Vector *make_f64vector(int size, double *eltp)
{
    return (ScmF64Vector*)Scm_MakeUVector(SCM_CLASS_F64VECTOR, size, eltp);
}

ScmObj Scm_MakeF64Vector(int size, double fill)
{
    ScmF64Vector *vec = make_f64vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = fill;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeF64VectorFromArray(int size, const double array[])
{
    ScmF64Vector *vec = make_f64vector(size, NULL);
    int i;
    for (i=0; i<size; i++) {
        vec->elements[i] = array[i];
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_MakeF64VectorFromArrayShared(int size, double array[])
{
    ScmF64Vector *vec = make_f64vector(size, array);
    return SCM_OBJ(vec);
}

ScmObj Scm_ListToF64Vector(ScmObj list, int clamp)
{
    int length = Scm_Length(list), i;
    ScmF64Vector *vec;
    ScmObj cp;

    if (length < 0) Scm_Error("improper list not allowed: %S", list);
    vec = make_f64vector(length, NULL);
    for (i=0, cp=list; i<length; i++, cp = SCM_CDR(cp)) {
        double elt;
        ScmObj obj = SCM_CAR(cp);
        elt = Scm_GetDouble(obj);
        vec->elements[i] = elt;
    }
    return SCM_OBJ(vec);
}

ScmObj Scm_VectorToF64Vector(ScmVector *ivec, int start, int end, int clamp)
{
    int length = SCM_VECTOR_SIZE(ivec), i;
    ScmF64Vector *vec;
    SCM_CHECK_START_END(start, end, length);
    vec = make_f64vector(end-start, NULL);
    for (i=start; i<end; i++) {
        double elt;
        ScmObj obj = SCM_VECTOR_ELEMENT(ivec, i);
        elt = Scm_GetDouble(obj);
        vec->elements[i-start] = elt;
    }
    return SCM_OBJ(vec);
}

/*
 * Accessors and modifiers
 */

ScmObj Scm_F64VectorFill(ScmF64Vector *vec, double fill, int start, int end)
{
    int i, size = SCM_F64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    for (i=start; i<end; i++) vec->elements[i] = fill;
    return SCM_OBJ(vec);
}

ScmObj Scm_F64VectorRef(ScmF64Vector *vec, int index, ScmObj fallback)
{
    ScmObj r;
    double elt;
    if (index < 0 || index >= SCM_F64VECTOR_SIZE(vec)) {
        if (SCM_UNBOUNDP(fallback)) 
            Scm_Error("index out of range: %d", index);
        return fallback;
    }
    elt = vec->elements[index];
    r = Scm_MakeFlonum(elt);
    return r;
}

ScmObj Scm_F64VectorSet(ScmF64Vector *vec, int index, ScmObj val, int clamp)
{
    double elt;
    if (index < 0 || index >= SCM_F64VECTOR_SIZE(vec))
        Scm_Error("index out of range: %d", index);
    SCM_UVECTOR_CHECK_MUTABLE(vec);
    elt = Scm_GetDouble(val);
    vec->elements[index] = elt;
    return SCM_OBJ(vec);
}

ScmObj Scm_F64VectorToList(ScmF64Vector *vec, int start, int end)
{
    ScmObj head = SCM_NIL, tail = SCM_NIL;
    int i, size = SCM_F64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    for (i=start; i<end; i++) {
        ScmObj obj;
        double elt = vec->elements[i];
        obj = Scm_MakeFlonum(elt);
        SCM_APPEND1(head, tail, obj);
    }
    return head;
}

ScmObj Scm_F64VectorToVector(ScmF64Vector *vec, int start, int end)
{
    ScmObj ovec;
    int i, size = SCM_F64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    ovec = Scm_MakeVector(end-start, SCM_UNDEFINED);
    for (i=start; i<end; i++) {
        ScmObj obj;
        double elt = vec->elements[i];
        obj = Scm_MakeFlonum(elt);
        SCM_VECTOR_ELEMENT(ovec, i-start) = obj;
    }
    return ovec;
}

ScmObj Scm_F64VectorCopy(ScmF64Vector *vec, int start, int end)
{
    int size = SCM_F64VECTOR_SIZE(vec);
    SCM_CHECK_START_END(start, end, size);
    return Scm_MakeF64VectorFromArray(end-start,
                                     SCM_F64VECTOR_ELEMENTS(vec)+start);
}

ScmObj Scm_F64VectorCopyX(ScmF64Vector *dst,
                           int dstart,
                           ScmF64Vector *src,
                           int sstart,
                           int send)
{
    int dlen = SCM_F64VECTOR_SIZE(dst);
    int slen = SCM_F64VECTOR_SIZE(src);
    int size;
    
    SCM_UVECTOR_CHECK_MUTABLE(dst);
    SCM_CHECK_START_END(sstart, send, slen);

    if (dstart < 0 || dstart >= dlen) return SCM_OBJ(dst);
    if (dlen - dstart > send - sstart)  size = send - sstart;
    else size = dlen - dstart;

    memcpy(SCM_F64VECTOR_ELEMENTS(dst) + dstart,
           SCM_F64VECTOR_ELEMENTS(src) + sstart,
           size * sizeof(double));
    return SCM_OBJ(dst);
}

static void s8vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S8VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S8VECTOR_ELEMENTS(s1)[i];
            r = s8s8_add(v0, v1, clamp);
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s8num(vv1, &oor);
            if (!oor) {
                r = s8g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s8num(vv1, &oor);
            if (!oor) {
                r = s8g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s8num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s8g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, s1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
    }
}

ScmObj Scm_S8VectorAdd(ScmS8Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S8VECTOR,
                               SCM_S8VECTOR_SIZE(s0),
                               NULL);
    s8vector_add("s8vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S8VectorAddX(ScmS8Vector *s0, ScmObj s1, int clamp)
{
    s8vector_add("s8vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u8vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U8VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U8VECTOR_ELEMENTS(s1)[i];
            r = u8u8_add(v0, v1, clamp);
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u8num(vv1, &oor);
            if (!oor) {
                r = u8g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u8num(vv1, &oor);
            if (!oor) {
                r = u8g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u8num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u8g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, s1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
    }
}

ScmObj Scm_U8VectorAdd(ScmU8Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U8VECTOR,
                               SCM_U8VECTOR_SIZE(s0),
                               NULL);
    u8vector_add("u8vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U8VectorAddX(ScmU8Vector *s0, ScmObj s1, int clamp)
{
    u8vector_add("u8vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s16vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S16VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S16VECTOR_ELEMENTS(s1)[i];
            r = s16s16_add(v0, v1, clamp);
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s16num(vv1, &oor);
            if (!oor) {
                r = s16g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s16num(vv1, &oor);
            if (!oor) {
                r = s16g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s16num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s16g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, s1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
    }
}

ScmObj Scm_S16VectorAdd(ScmS16Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S16VECTOR,
                               SCM_S16VECTOR_SIZE(s0),
                               NULL);
    s16vector_add("s16vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S16VectorAddX(ScmS16Vector *s0, ScmObj s1, int clamp)
{
    s16vector_add("s16vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u16vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U16VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U16VECTOR_ELEMENTS(s1)[i];
            r = u16u16_add(v0, v1, clamp);
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u16num(vv1, &oor);
            if (!oor) {
                r = u16g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u16num(vv1, &oor);
            if (!oor) {
                r = u16g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, vv1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u16num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u16g_add(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Add(rr, s1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
    }
}

ScmObj Scm_U16VectorAdd(ScmU16Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U16VECTOR,
                               SCM_U16VECTOR_SIZE(s0),
                               NULL);
    u16vector_add("u16vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U16VectorAddX(ScmU16Vector *s0, ScmObj s1, int clamp)
{
    u16vector_add("u16vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s32vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S32VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S32VECTOR_ELEMENTS(s1)[i];
            r = s32s32_add(v0, v1, clamp);
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s32num(vv1, &oor);
            if (!oor) {
                r = s32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s32num(vv1, &oor);
            if (!oor) {
                r = s32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Add(rr, s1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
    }
}

ScmObj Scm_S32VectorAdd(ScmS32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S32VECTOR,
                               SCM_S32VECTOR_SIZE(s0),
                               NULL);
    s32vector_add("s32vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S32VectorAddX(ScmS32Vector *s0, ScmObj s1, int clamp)
{
    s32vector_add("s32vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u32vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U32VECTOR_SIZE(d), oor;
    u_long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U32VECTOR_ELEMENTS(s1)[i];
            r = u32u32_add(v0, v1, clamp);
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u32num(vv1, &oor);
            if (!oor) {
                r = u32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u32num(vv1, &oor);
            if (!oor) {
                r = u32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Add(rr, s1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
    }
}

ScmObj Scm_U32VectorAdd(ScmU32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U32VECTOR,
                               SCM_U32VECTOR_SIZE(s0),
                               NULL);
    u32vector_add("u32vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U32VectorAddX(ScmU32Vector *s0, ScmObj s1, int clamp)
{
    u32vector_add("u32vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s64vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S64VECTOR_SIZE(d), oor;
    ScmInt64 r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S64VECTOR_ELEMENTS(s1)[i];
            r = s64s64_add(v0, v1, clamp);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s64num(vv1, &oor);
            if (!oor) {
                r = s64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s64num(vv1, &oor);
            if (!oor) {
                r = s64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Add(rr, s1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
    }
}

ScmObj Scm_S64VectorAdd(ScmS64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S64VECTOR,
                               SCM_S64VECTOR_SIZE(s0),
                               NULL);
    s64vector_add("s64vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S64VectorAddX(ScmS64Vector *s0, ScmObj s1, int clamp)
{
    s64vector_add("s64vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u64vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U64VECTOR_SIZE(d), oor;
    ScmUInt64 r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U64VECTOR_ELEMENTS(s1)[i];
            r = u64u64_add(v0, v1, clamp);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u64num(vv1, &oor);
            if (!oor) {
                r = u64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u64num(vv1, &oor);
            if (!oor) {
                r = u64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Add(rr, s1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
    }
}

ScmObj Scm_U64VectorAdd(ScmU64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U64VECTOR,
                               SCM_U64VECTOR_SIZE(s0),
                               NULL);
    u64vector_add("u64vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U64VectorAddX(ScmU64Vector *s0, ScmObj s1, int clamp)
{
    u64vector_add("u64vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f32vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F32VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F32VECTOR_ELEMENTS(s1)[i];
            r = f32f32_add(v0, v1, clamp);
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Add(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Add(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f32g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Add(rr, s1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
    }
}

ScmObj Scm_F32VectorAdd(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F32VECTOR,
                               SCM_F32VECTOR_SIZE(s0),
                               NULL);
    f32vector_add("f32vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F32VectorAddX(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    f32vector_add("f32vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f64vector_add(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F64VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F64VECTOR_ELEMENTS(s1)[i];
            r = f64f64_add(v0, v1, clamp);
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Add(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f64g_add(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Add(rr, s1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
    }
}

ScmObj Scm_F64VectorAdd(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F64VECTOR,
                               SCM_F64VECTOR_SIZE(s0),
                               NULL);
    f64vector_add("f64vector-add", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F64VectorAddX(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    f64vector_add("f64vector-add!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s8vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S8VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S8VECTOR_ELEMENTS(s1)[i];
            r = s8s8_sub(v0, v1, clamp);
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s8num(vv1, &oor);
            if (!oor) {
                r = s8g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s8num(vv1, &oor);
            if (!oor) {
                r = s8g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s8num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s8g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, s1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
    }
}

ScmObj Scm_S8VectorSub(ScmS8Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S8VECTOR,
                               SCM_S8VECTOR_SIZE(s0),
                               NULL);
    s8vector_sub("s8vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S8VectorSubX(ScmS8Vector *s0, ScmObj s1, int clamp)
{
    s8vector_sub("s8vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u8vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U8VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U8VECTOR_ELEMENTS(s1)[i];
            r = u8u8_sub(v0, v1, clamp);
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u8num(vv1, &oor);
            if (!oor) {
                r = u8g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u8num(vv1, &oor);
            if (!oor) {
                r = u8g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u8num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u8g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, s1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
    }
}

ScmObj Scm_U8VectorSub(ScmU8Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U8VECTOR,
                               SCM_U8VECTOR_SIZE(s0),
                               NULL);
    u8vector_sub("u8vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U8VectorSubX(ScmU8Vector *s0, ScmObj s1, int clamp)
{
    u8vector_sub("u8vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s16vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S16VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S16VECTOR_ELEMENTS(s1)[i];
            r = s16s16_sub(v0, v1, clamp);
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s16num(vv1, &oor);
            if (!oor) {
                r = s16g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s16num(vv1, &oor);
            if (!oor) {
                r = s16g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s16num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s16g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, s1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
    }
}

ScmObj Scm_S16VectorSub(ScmS16Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S16VECTOR,
                               SCM_S16VECTOR_SIZE(s0),
                               NULL);
    s16vector_sub("s16vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S16VectorSubX(ScmS16Vector *s0, ScmObj s1, int clamp)
{
    s16vector_sub("s16vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u16vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U16VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U16VECTOR_ELEMENTS(s1)[i];
            r = u16u16_sub(v0, v1, clamp);
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u16num(vv1, &oor);
            if (!oor) {
                r = u16g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u16num(vv1, &oor);
            if (!oor) {
                r = u16g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, vv1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u16num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u16g_sub(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Sub(rr, s1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
    }
}

ScmObj Scm_U16VectorSub(ScmU16Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U16VECTOR,
                               SCM_U16VECTOR_SIZE(s0),
                               NULL);
    u16vector_sub("u16vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U16VectorSubX(ScmU16Vector *s0, ScmObj s1, int clamp)
{
    u16vector_sub("u16vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s32vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S32VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S32VECTOR_ELEMENTS(s1)[i];
            r = s32s32_sub(v0, v1, clamp);
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s32num(vv1, &oor);
            if (!oor) {
                r = s32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s32num(vv1, &oor);
            if (!oor) {
                r = s32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Sub(rr, s1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
    }
}

ScmObj Scm_S32VectorSub(ScmS32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S32VECTOR,
                               SCM_S32VECTOR_SIZE(s0),
                               NULL);
    s32vector_sub("s32vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S32VectorSubX(ScmS32Vector *s0, ScmObj s1, int clamp)
{
    s32vector_sub("s32vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u32vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U32VECTOR_SIZE(d), oor;
    u_long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U32VECTOR_ELEMENTS(s1)[i];
            r = u32u32_sub(v0, v1, clamp);
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u32num(vv1, &oor);
            if (!oor) {
                r = u32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u32num(vv1, &oor);
            if (!oor) {
                r = u32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Sub(rr, s1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
    }
}

ScmObj Scm_U32VectorSub(ScmU32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U32VECTOR,
                               SCM_U32VECTOR_SIZE(s0),
                               NULL);
    u32vector_sub("u32vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U32VectorSubX(ScmU32Vector *s0, ScmObj s1, int clamp)
{
    u32vector_sub("u32vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s64vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S64VECTOR_SIZE(d), oor;
    ScmInt64 r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S64VECTOR_ELEMENTS(s1)[i];
            r = s64s64_sub(v0, v1, clamp);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s64num(vv1, &oor);
            if (!oor) {
                r = s64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s64num(vv1, &oor);
            if (!oor) {
                r = s64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Sub(rr, s1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
    }
}

ScmObj Scm_S64VectorSub(ScmS64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S64VECTOR,
                               SCM_S64VECTOR_SIZE(s0),
                               NULL);
    s64vector_sub("s64vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S64VectorSubX(ScmS64Vector *s0, ScmObj s1, int clamp)
{
    s64vector_sub("s64vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u64vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U64VECTOR_SIZE(d), oor;
    ScmUInt64 r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U64VECTOR_ELEMENTS(s1)[i];
            r = u64u64_sub(v0, v1, clamp);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u64num(vv1, &oor);
            if (!oor) {
                r = u64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u64num(vv1, &oor);
            if (!oor) {
                r = u64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Sub(rr, s1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
    }
}

ScmObj Scm_U64VectorSub(ScmU64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U64VECTOR,
                               SCM_U64VECTOR_SIZE(s0),
                               NULL);
    u64vector_sub("u64vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U64VectorSubX(ScmU64Vector *s0, ScmObj s1, int clamp)
{
    u64vector_sub("u64vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f32vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F32VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F32VECTOR_ELEMENTS(s1)[i];
            r = f32f32_sub(v0, v1, clamp);
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Sub(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Sub(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f32g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Sub(rr, s1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
    }
}

ScmObj Scm_F32VectorSub(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F32VECTOR,
                               SCM_F32VECTOR_SIZE(s0),
                               NULL);
    f32vector_sub("f32vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F32VectorSubX(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    f32vector_sub("f32vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f64vector_sub(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F64VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F64VECTOR_ELEMENTS(s1)[i];
            r = f64f64_sub(v0, v1, clamp);
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Sub(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f64g_sub(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Sub(rr, s1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
    }
}

ScmObj Scm_F64VectorSub(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F64VECTOR,
                               SCM_F64VECTOR_SIZE(s0),
                               NULL);
    f64vector_sub("f64vector-sub", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F64VectorSubX(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    f64vector_sub("f64vector-sub!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s8vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S8VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S8VECTOR_ELEMENTS(s1)[i];
            r = s8s8_mul(v0, v1, clamp);
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s8num(vv1, &oor);
            if (!oor) {
                r = s8g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s8num(vv1, &oor);
            if (!oor) {
                r = s8g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s8num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s8g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, s1);
                r = s8unbox(rr, clamp);
            }
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
    }
}

ScmObj Scm_S8VectorMul(ScmS8Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S8VECTOR,
                               SCM_S8VECTOR_SIZE(s0),
                               NULL);
    s8vector_mul("s8vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S8VectorMulX(ScmS8Vector *s0, ScmObj s1, int clamp)
{
    s8vector_mul("s8vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u8vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U8VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U8VECTOR_ELEMENTS(s1)[i];
            r = u8u8_mul(v0, v1, clamp);
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u8num(vv1, &oor);
            if (!oor) {
                r = u8g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u8num(vv1, &oor);
            if (!oor) {
                r = u8g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u8num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u8g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, s1);
                r = u8unbox(rr, clamp);
            }
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
    }
}

ScmObj Scm_U8VectorMul(ScmU8Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U8VECTOR,
                               SCM_U8VECTOR_SIZE(s0),
                               NULL);
    u8vector_mul("u8vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U8VectorMulX(ScmU8Vector *s0, ScmObj s1, int clamp)
{
    u8vector_mul("u8vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s16vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S16VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S16VECTOR_ELEMENTS(s1)[i];
            r = s16s16_mul(v0, v1, clamp);
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s16num(vv1, &oor);
            if (!oor) {
                r = s16g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s16num(vv1, &oor);
            if (!oor) {
                r = s16g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s16num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s16g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, s1);
                r = s16unbox(rr, clamp);
            }
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
    }
}

ScmObj Scm_S16VectorMul(ScmS16Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S16VECTOR,
                               SCM_S16VECTOR_SIZE(s0),
                               NULL);
    s16vector_mul("s16vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S16VectorMulX(ScmS16Vector *s0, ScmObj s1, int clamp)
{
    s16vector_mul("s16vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u16vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U16VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U16VECTOR_ELEMENTS(s1)[i];
            r = u16u16_mul(v0, v1, clamp);
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u16num(vv1, &oor);
            if (!oor) {
                r = u16g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u16num(vv1, &oor);
            if (!oor) {
                r = u16g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, vv1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u16num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u16g_mul(v0, v1, clamp);
            } else {
                rr = SCM_MAKE_INT(v0);
                rr = Scm_Mul(rr, s1);
                r = u16unbox(rr, clamp);
            }
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
    }
}

ScmObj Scm_U16VectorMul(ScmU16Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U16VECTOR,
                               SCM_U16VECTOR_SIZE(s0),
                               NULL);
    u16vector_mul("u16vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U16VectorMulX(ScmU16Vector *s0, ScmObj s1, int clamp)
{
    u16vector_mul("u16vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s32vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S32VECTOR_SIZE(d), oor;
    long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S32VECTOR_ELEMENTS(s1)[i];
            r = s32s32_mul(v0, v1, clamp);
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s32num(vv1, &oor);
            if (!oor) {
                r = s32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s32num(vv1, &oor);
            if (!oor) {
                r = s32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger(v0);
                rr = Scm_Mul(rr, s1);
                r = Scm_GetInteger32Clamp(rr, clamp, NULL);
            }
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
    }
}

ScmObj Scm_S32VectorMul(ScmS32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S32VECTOR,
                               SCM_S32VECTOR_SIZE(s0),
                               NULL);
    s32vector_mul("s32vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S32VectorMulX(ScmS32Vector *s0, ScmObj s1, int clamp)
{
    s32vector_mul("s32vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u32vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U32VECTOR_SIZE(d), oor;
    u_long r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U32VECTOR_ELEMENTS(s1)[i];
            r = u32u32_mul(v0, v1, clamp);
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u32num(vv1, &oor);
            if (!oor) {
                r = u32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u32num(vv1, &oor);
            if (!oor) {
                r = u32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU(v0);
                rr = Scm_Mul(rr, s1);
                r = Scm_GetIntegerU32Clamp(rr, clamp, NULL);
            }
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
    }
}

ScmObj Scm_U32VectorMul(ScmU32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U32VECTOR,
                               SCM_U32VECTOR_SIZE(s0),
                               NULL);
    u32vector_mul("u32vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U32VectorMulX(ScmU32Vector *s0, ScmObj s1, int clamp)
{
    u32vector_mul("u32vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s64vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_S64VECTOR_SIZE(d), oor;
    ScmInt64 r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S64VECTOR_ELEMENTS(s1)[i];
            r = s64s64_mul(v0, v1, clamp);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = s64num(vv1, &oor);
            if (!oor) {
                r = s64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = s64num(vv1, &oor);
            if (!oor) {
                r = s64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = s64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = s64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeInteger64(v0);
                rr = Scm_Mul(rr, s1);
                r = Scm_GetInteger64Clamp(rr, clamp, NULL);
            }
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
    }
}

ScmObj Scm_S64VectorMul(ScmS64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S64VECTOR,
                               SCM_S64VECTOR_SIZE(s0),
                               NULL);
    s64vector_mul("s64vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_S64VectorMulX(ScmS64Vector *s0, ScmObj s1, int clamp)
{
    s64vector_mul("s64vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void u64vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_U64VECTOR_SIZE(d), oor;
    ScmUInt64 r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U64VECTOR_ELEMENTS(s1)[i];
            r = u64u64_mul(v0, v1, clamp);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = u64num(vv1, &oor);
            if (!oor) {
                r = u64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = u64num(vv1, &oor);
            if (!oor) {
                r = u64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = u64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = u64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeIntegerU64(v0);
                rr = Scm_Mul(rr, s1);
                r = Scm_GetIntegerU64Clamp(rr, clamp, NULL);
            }
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
    }
}

ScmObj Scm_U64VectorMul(ScmU64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U64VECTOR,
                               SCM_U64VECTOR_SIZE(s0),
                               NULL);
    u64vector_mul("u64vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_U64VectorMulX(ScmU64Vector *s0, ScmObj s1, int clamp)
{
    u64vector_mul("u64vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f32vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F32VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F32VECTOR_ELEMENTS(s1)[i];
            r = f32f32_mul(v0, v1, clamp);
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Mul(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Mul(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f32g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Mul(rr, s1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
    }
}

ScmObj Scm_F32VectorMul(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F32VECTOR,
                               SCM_F32VECTOR_SIZE(s0),
                               NULL);
    f32vector_mul("f32vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F32VectorMulX(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    f32vector_mul("f32vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f64vector_mul(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F64VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F64VECTOR_ELEMENTS(s1)[i];
            r = f64f64_mul(v0, v1, clamp);
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Mul(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f64g_mul(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Mul(rr, s1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
    }
}

ScmObj Scm_F64VectorMul(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F64VECTOR,
                               SCM_F64VECTOR_SIZE(s0),
                               NULL);
    f64vector_mul("f64vector-mul", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F64VectorMulX(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    f64vector_mul("f64vector-mul!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f32vector_div(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F32VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F32VECTOR_ELEMENTS(s1)[i];
            r = f32f32_div(v0, v1, clamp);
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_div(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Div(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f32num(vv1, &oor);
            if (!oor) {
                r = f32g_div(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Div(rr, vv1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f32num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F32VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f32g_div(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum((double)v0);
                rr = Scm_Div(rr, s1);
                r = (float)Scm_GetDouble(rr);
            }
            SCM_F32VECTOR_ELEMENTS(d)[i] = (float)r;
        }
    }
}

ScmObj Scm_F32VectorDiv(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F32VECTOR,
                               SCM_F32VECTOR_SIZE(s0),
                               NULL);
    f32vector_div("f32vector-div", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F32VectorDivX(ScmF32Vector *s0, ScmObj s1, int clamp)
{
    f32vector_div("f32vector-div!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void f64vector_div(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1, int clamp)
{
    int i, size = SCM_F64VECTOR_SIZE(d), oor;
    double r, v0, v1;
    ScmObj rr, vv1;

    switch (arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_F64VECTOR_ELEMENTS(s1)[i];
            r = f64f64_div(v0, v1, clamp);
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_div(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Div(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_CAR(s1); s1 = SCM_CDR(s1);
            v1 = f64num(vv1, &oor);
            if (!oor) {
                r = f64g_div(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Div(rr, vv1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = f64num(s1, &oor);
        for (i=0; i<size; i++) {
            v0 = SCM_F64VECTOR_ELEMENTS(s0)[i];
            if (!oor) {
                r = f64g_div(v0, v1, clamp);
            } else {
                rr = Scm_MakeFlonum(v0);
                rr = Scm_Div(rr, s1);
                r = Scm_GetDouble(rr);
            }
            SCM_F64VECTOR_ELEMENTS(d)[i] = (double)r;
        }
    }
}

ScmObj Scm_F64VectorDiv(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_F64VECTOR,
                               SCM_F64VECTOR_SIZE(s0),
                               NULL);
    f64vector_div("f64vector-div", d, SCM_OBJ(s0), s1, clamp);
    return d;
}

ScmObj Scm_F64VectorDivX(ScmF64Vector *s0, ScmObj s1, int clamp)
{
    f64vector_div("f64vector-div!", SCM_OBJ(s0), SCM_OBJ(s0), s1, clamp);
    return SCM_OBJ(s0);
}

static void s8vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S8VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S8VECTOR_ELEMENTS(s1)[i];
            r = v0 & v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            r = v0 & v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
    }
}

ScmObj Scm_S8VectorAnd(ScmS8Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S8VECTOR,
                               SCM_S8VECTOR_SIZE(s0),
                               NULL);
    s8vector_and("s8vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S8VectorAndX(ScmS8Vector *s0, ScmObj s1)
{
    s8vector_and("s8vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s8vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S8VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S8VECTOR_ELEMENTS(s1)[i];
            r = v0 | v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            r = v0 | v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
    }
}

ScmObj Scm_S8VectorIor(ScmS8Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S8VECTOR,
                               SCM_S8VECTOR_SIZE(s0),
                               NULL);
    s8vector_ior("s8vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S8VectorIorX(ScmS8Vector *s0, ScmObj s1)
{
    s8vector_ior("s8vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s8vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S8VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S8VECTOR_ELEMENTS(s1)[i];
            r = v0 ^ v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S8VECTOR_ELEMENTS(s0)[i];
            r = v0 ^ v1;
            SCM_S8VECTOR_ELEMENTS(d)[i] = (signed char)r;
        }
    }
}

ScmObj Scm_S8VectorXor(ScmS8Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S8VECTOR,
                               SCM_S8VECTOR_SIZE(s0),
                               NULL);
    s8vector_xor("s8vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S8VectorXorX(ScmS8Vector *s0, ScmObj s1)
{
    s8vector_xor("s8vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u8vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U8VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U8VECTOR_ELEMENTS(s1)[i];
            r = v0 & v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            r = v0 & v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
    }
}

ScmObj Scm_U8VectorAnd(ScmU8Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U8VECTOR,
                               SCM_U8VECTOR_SIZE(s0),
                               NULL);
    u8vector_and("u8vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U8VectorAndX(ScmU8Vector *s0, ScmObj s1)
{
    u8vector_and("u8vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u8vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U8VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U8VECTOR_ELEMENTS(s1)[i];
            r = v0 | v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            r = v0 | v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
    }
}

ScmObj Scm_U8VectorIor(ScmU8Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U8VECTOR,
                               SCM_U8VECTOR_SIZE(s0),
                               NULL);
    u8vector_ior("u8vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U8VectorIorX(ScmU8Vector *s0, ScmObj s1)
{
    u8vector_ior("u8vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u8vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U8VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U8VECTOR_ELEMENTS(s1)[i];
            r = v0 ^ v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U8VECTOR_ELEMENTS(s0)[i];
            r = v0 ^ v1;
            SCM_U8VECTOR_ELEMENTS(d)[i] = (unsigned char)r;
        }
    }
}

ScmObj Scm_U8VectorXor(ScmU8Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U8VECTOR,
                               SCM_U8VECTOR_SIZE(s0),
                               NULL);
    u8vector_xor("u8vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U8VectorXorX(ScmU8Vector *s0, ScmObj s1)
{
    u8vector_xor("u8vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s16vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S16VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S16VECTOR_ELEMENTS(s1)[i];
            r = v0 & v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            r = v0 & v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
    }
}

ScmObj Scm_S16VectorAnd(ScmS16Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S16VECTOR,
                               SCM_S16VECTOR_SIZE(s0),
                               NULL);
    s16vector_and("s16vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S16VectorAndX(ScmS16Vector *s0, ScmObj s1)
{
    s16vector_and("s16vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s16vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S16VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S16VECTOR_ELEMENTS(s1)[i];
            r = v0 | v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            r = v0 | v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
    }
}

ScmObj Scm_S16VectorIor(ScmS16Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S16VECTOR,
                               SCM_S16VECTOR_SIZE(s0),
                               NULL);
    s16vector_ior("s16vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S16VectorIorX(ScmS16Vector *s0, ScmObj s1)
{
    s16vector_ior("s16vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s16vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S16VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S16VECTOR_ELEMENTS(s1)[i];
            r = v0 ^ v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S16VECTOR_ELEMENTS(s0)[i];
            r = v0 ^ v1;
            SCM_S16VECTOR_ELEMENTS(d)[i] = (short)r;
        }
    }
}

ScmObj Scm_S16VectorXor(ScmS16Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S16VECTOR,
                               SCM_S16VECTOR_SIZE(s0),
                               NULL);
    s16vector_xor("s16vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S16VectorXorX(ScmS16Vector *s0, ScmObj s1)
{
    s16vector_xor("s16vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u16vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U16VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U16VECTOR_ELEMENTS(s1)[i];
            r = v0 & v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            r = v0 & v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
    }
}

ScmObj Scm_U16VectorAnd(ScmU16Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U16VECTOR,
                               SCM_U16VECTOR_SIZE(s0),
                               NULL);
    u16vector_and("u16vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U16VectorAndX(ScmU16Vector *s0, ScmObj s1)
{
    u16vector_and("u16vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u16vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U16VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U16VECTOR_ELEMENTS(s1)[i];
            r = v0 | v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            r = v0 | v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
    }
}

ScmObj Scm_U16VectorIor(ScmU16Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U16VECTOR,
                               SCM_U16VECTOR_SIZE(s0),
                               NULL);
    u16vector_ior("u16vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U16VectorIorX(ScmU16Vector *s0, ScmObj s1)
{
    u16vector_ior("u16vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u16vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U16VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U16VECTOR_ELEMENTS(s1)[i];
            r = v0 ^ v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U16VECTOR_ELEMENTS(s0)[i];
            r = v0 ^ v1;
            SCM_U16VECTOR_ELEMENTS(d)[i] = (unsigned short)r;
        }
    }
}

ScmObj Scm_U16VectorXor(ScmU16Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U16VECTOR,
                               SCM_U16VECTOR_SIZE(s0),
                               NULL);
    u16vector_xor("u16vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U16VectorXorX(ScmU16Vector *s0, ScmObj s1)
{
    u16vector_xor("u16vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s32vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S32VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S32VECTOR_ELEMENTS(s1)[i];
            r = v0 & v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            r = v0 & v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
    }
}

ScmObj Scm_S32VectorAnd(ScmS32Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S32VECTOR,
                               SCM_S32VECTOR_SIZE(s0),
                               NULL);
    s32vector_and("s32vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S32VectorAndX(ScmS32Vector *s0, ScmObj s1)
{
    s32vector_and("s32vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s32vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S32VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S32VECTOR_ELEMENTS(s1)[i];
            r = v0 | v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            r = v0 | v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
    }
}

ScmObj Scm_S32VectorIor(ScmS32Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S32VECTOR,
                               SCM_S32VECTOR_SIZE(s0),
                               NULL);
    s32vector_ior("s32vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S32VectorIorX(ScmS32Vector *s0, ScmObj s1)
{
    s32vector_ior("s32vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s32vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S32VECTOR_SIZE(d);
    long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S32VECTOR_ELEMENTS(s1)[i];
            r = v0 ^ v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S32VECTOR_ELEMENTS(s0)[i];
            r = v0 ^ v1;
            SCM_S32VECTOR_ELEMENTS(d)[i] = (ScmInt32)r;
        }
    }
}

ScmObj Scm_S32VectorXor(ScmS32Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S32VECTOR,
                               SCM_S32VECTOR_SIZE(s0),
                               NULL);
    s32vector_xor("s32vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S32VectorXorX(ScmS32Vector *s0, ScmObj s1)
{
    s32vector_xor("s32vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u32vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U32VECTOR_SIZE(d);
    u_long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U32VECTOR_ELEMENTS(s1)[i];
            r = v0 & v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 & v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            r = v0 & v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
    }
}

ScmObj Scm_U32VectorAnd(ScmU32Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U32VECTOR,
                               SCM_U32VECTOR_SIZE(s0),
                               NULL);
    u32vector_and("u32vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U32VectorAndX(ScmU32Vector *s0, ScmObj s1)
{
    u32vector_and("u32vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u32vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U32VECTOR_SIZE(d);
    u_long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U32VECTOR_ELEMENTS(s1)[i];
            r = v0 | v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 | v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            r = v0 | v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
    }
}

ScmObj Scm_U32VectorIor(ScmU32Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U32VECTOR,
                               SCM_U32VECTOR_SIZE(s0),
                               NULL);
    u32vector_ior("u32vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U32VectorIorX(ScmU32Vector *s0, ScmObj s1)
{
    u32vector_ior("u32vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u32vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U32VECTOR_SIZE(d);
    u_long r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U32VECTOR_ELEMENTS(s1)[i];
            r = v0 ^ v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext(vv1);
            r = v0 ^ v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U32VECTOR_ELEMENTS(s0)[i];
            r = v0 ^ v1;
            SCM_U32VECTOR_ELEMENTS(d)[i] = (ScmUInt32)r;
        }
    }
}

ScmObj Scm_U32VectorXor(ScmU32Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U32VECTOR,
                               SCM_U32VECTOR_SIZE(s0),
                               NULL);
    u32vector_xor("u32vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U32VectorXorX(ScmU32Vector *s0, ScmObj s1)
{
    u32vector_xor("u32vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s64vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S64VECTOR_SIZE(d);
    ScmInt64 r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S64VECTOR_ELEMENTS(s1)[i];
            INT64BITOP(r, v0, &, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, &, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, &, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext64(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            INT64BITOP(r, v0, &, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
    }
}

ScmObj Scm_S64VectorAnd(ScmS64Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S64VECTOR,
                               SCM_S64VECTOR_SIZE(s0),
                               NULL);
    s64vector_and("s64vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S64VectorAndX(ScmS64Vector *s0, ScmObj s1)
{
    s64vector_and("s64vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s64vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S64VECTOR_SIZE(d);
    ScmInt64 r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S64VECTOR_ELEMENTS(s1)[i];
            INT64BITOP(r, v0, |, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, |, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, |, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext64(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            INT64BITOP(r, v0, |, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
    }
}

ScmObj Scm_S64VectorIor(ScmS64Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S64VECTOR,
                               SCM_S64VECTOR_SIZE(s0),
                               NULL);
    s64vector_ior("s64vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S64VectorIorX(ScmS64Vector *s0, ScmObj s1)
{
    s64vector_ior("s64vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void s64vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_S64VECTOR_SIZE(d);
    ScmInt64 r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_S64VECTOR_ELEMENTS(s1)[i];
            INT64BITOP(r, v0, ^, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, ^, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, ^, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext64(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_S64VECTOR_ELEMENTS(s0)[i];
            INT64BITOP(r, v0, ^, v1);
            SCM_S64VECTOR_ELEMENTS(d)[i] = (ScmInt64)r;
        }
    }
}

ScmObj Scm_S64VectorXor(ScmS64Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_S64VECTOR,
                               SCM_S64VECTOR_SIZE(s0),
                               NULL);
    s64vector_xor("s64vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_S64VectorXorX(ScmS64Vector *s0, ScmObj s1)
{
    s64vector_xor("s64vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u64vector_and(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U64VECTOR_SIZE(d);
    ScmUInt64 r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U64VECTOR_ELEMENTS(s1)[i];
            INT64BITOP(r, v0, &, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, &, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, &, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext64(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            INT64BITOP(r, v0, &, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
    }
}

ScmObj Scm_U64VectorAnd(ScmU64Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U64VECTOR,
                               SCM_U64VECTOR_SIZE(s0),
                               NULL);
    u64vector_and("u64vector-and", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U64VectorAndX(ScmU64Vector *s0, ScmObj s1)
{
    u64vector_and("u64vector-and!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u64vector_ior(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U64VECTOR_SIZE(d);
    ScmUInt64 r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U64VECTOR_ELEMENTS(s1)[i];
            INT64BITOP(r, v0, |, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, |, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, |, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext64(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            INT64BITOP(r, v0, |, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
    }
}

ScmObj Scm_U64VectorIor(ScmU64Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U64VECTOR,
                               SCM_U64VECTOR_SIZE(s0),
                               NULL);
    u64vector_ior("u64vector-ior", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U64VectorIorX(ScmU64Vector *s0, ScmObj s1)
{
    u64vector_ior("u64vector-ior!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
static void u64vector_xor(const char *name,
                                 ScmObj d, ScmObj s0, ScmObj s1)
{
    int i, size = SCM_U64VECTOR_SIZE(d);
    ScmUInt64 r, v0, v1;
    ScmObj vv1;
    
    switch(arg2_check(name, s0, s1, TRUE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            v1 = SCM_U64VECTOR_ELEMENTS(s1)[i];
            INT64BITOP(r, v0, ^, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, ^, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            vv1 = SCM_VECTOR_ELEMENTS(s1)[i];
            v1 = bitext64(vv1);
            INT64BITOP(r, v0, ^, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
        break;
    case ARGTYPE_CONST:
        v1 = bitext64(s1);
        for (i=0; i<size; i++) {
            v0 = SCM_U64VECTOR_ELEMENTS(s0)[i];
            INT64BITOP(r, v0, ^, v1);
            SCM_U64VECTOR_ELEMENTS(d)[i] = (ScmUInt64)r;
        }
    }
}

ScmObj Scm_U64VectorXor(ScmU64Vector *s0, ScmObj s1)
{
    ScmObj d = Scm_MakeUVector(SCM_CLASS_U64VECTOR,
                               SCM_U64VECTOR_SIZE(s0),
                               NULL);
    u64vector_xor("u64vector-xor", d, SCM_OBJ(s0), s1);
    return d;
}

ScmObj Scm_U64VectorXorX(ScmU64Vector *s0, ScmObj s1)
{
    u64vector_xor("u64vector-xor!", SCM_OBJ(s0), SCM_OBJ(s0), s1);
    return SCM_OBJ(s0);
}
ScmObj Scm_S8VectorDotProd(ScmS8Vector *x, ScmObj y)
{
    int i, size = SCM_S8VECTOR_SIZE(x), oor;
    long r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("s8vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S8VECTOR_ELEMENTS(x)[i];
            vy = SCM_S8VECTOR_ELEMENTS(y)[i];
            r = s8muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S8VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = s8num(vvy, &oor);
            if (!oor) {
                r = s8muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_S8VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = s8num(vvy, &oor);
            if (!oor) {
                r = s8muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of s8vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeInteger(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeInteger(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_U8VectorDotProd(ScmU8Vector *x, ScmObj y)
{
    int i, size = SCM_U8VECTOR_SIZE(x), oor;
    long r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("u8vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U8VECTOR_ELEMENTS(x)[i];
            vy = SCM_U8VECTOR_ELEMENTS(y)[i];
            r = u8muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U8VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = u8num(vvy, &oor);
            if (!oor) {
                r = u8muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_U8VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = u8num(vvy, &oor);
            if (!oor) {
                r = u8muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of u8vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeIntegerU(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeIntegerU(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_S16VectorDotProd(ScmS16Vector *x, ScmObj y)
{
    int i, size = SCM_S16VECTOR_SIZE(x), oor;
    long r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("s16vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S16VECTOR_ELEMENTS(x)[i];
            vy = SCM_S16VECTOR_ELEMENTS(y)[i];
            r = s16muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S16VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = s16num(vvy, &oor);
            if (!oor) {
                r = s16muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_S16VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = s16num(vvy, &oor);
            if (!oor) {
                r = s16muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of s16vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeInteger(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeInteger(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_U16VectorDotProd(ScmU16Vector *x, ScmObj y)
{
    int i, size = SCM_U16VECTOR_SIZE(x), oor;
    long r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("u16vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U16VECTOR_ELEMENTS(x)[i];
            vy = SCM_U16VECTOR_ELEMENTS(y)[i];
            r = u16muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U16VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = u16num(vvy, &oor);
            if (!oor) {
                r = u16muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_U16VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = u16num(vvy, &oor);
            if (!oor) {
                r = u16muladd(vx, vy, r, &rr);
            } else {
                vvx = SCM_MAKE_INT(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of u16vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeIntegerU(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeIntegerU(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_S32VectorDotProd(ScmS32Vector *x, ScmObj y)
{
    int i, size = SCM_S32VECTOR_SIZE(x), oor;
    long r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("s32vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S32VECTOR_ELEMENTS(x)[i];
            vy = SCM_S32VECTOR_ELEMENTS(y)[i];
            r = s32muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S32VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = s32num(vvy, &oor);
            if (!oor) {
                r = s32muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeInteger(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_S32VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = s32num(vvy, &oor);
            if (!oor) {
                r = s32muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeInteger(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of s32vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeInteger(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeInteger(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_U32VectorDotProd(ScmU32Vector *x, ScmObj y)
{
    int i, size = SCM_U32VECTOR_SIZE(x), oor;
    u_long r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("u32vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U32VECTOR_ELEMENTS(x)[i];
            vy = SCM_U32VECTOR_ELEMENTS(y)[i];
            r = u32muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U32VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = u32num(vvy, &oor);
            if (!oor) {
                r = u32muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeIntegerU(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_U32VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = u32num(vvy, &oor);
            if (!oor) {
                r = u32muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeIntegerU(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of u32vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeIntegerU(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeIntegerU(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_S64VectorDotProd(ScmS64Vector *x, ScmObj y)
{
    int i, size = SCM_S64VECTOR_SIZE(x), oor;
    ScmInt64 r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    SCM_SET_INT64_ZERO(r);
    switch (arg2_check("s64vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S64VECTOR_ELEMENTS(x)[i];
            vy = SCM_S64VECTOR_ELEMENTS(y)[i];
            r = s64muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_S64VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = s64num(vvy, &oor);
            if (!oor) {
                r = s64muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeInteger64(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_S64VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = s64num(vvy, &oor);
            if (!oor) {
                r = s64muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeInteger64(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of s64vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeInteger64(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeInteger64(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_U64VectorDotProd(ScmU64Vector *x, ScmObj y)
{
    int i, size = SCM_U64VECTOR_SIZE(x), oor;
    ScmUInt64 r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    SCM_SET_INT64_ZERO(r);
    switch (arg2_check("u64vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U64VECTOR_ELEMENTS(x)[i];
            vy = SCM_U64VECTOR_ELEMENTS(y)[i];
            r = u64muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_U64VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = u64num(vvy, &oor);
            if (!oor) {
                r = u64muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeIntegerU64(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_U64VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = u64num(vvy, &oor);
            if (!oor) {
                r = u64muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeIntegerU64(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of u64vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeIntegerU64(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeIntegerU64(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_F32VectorDotProd(ScmF32Vector *x, ScmObj y)
{
    int i, size = SCM_F32VECTOR_SIZE(x), oor;
    double r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("f32vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_F32VECTOR_ELEMENTS(x)[i];
            vy = SCM_F32VECTOR_ELEMENTS(y)[i];
            r = f32muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_F32VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = f32num(vvy, &oor);
            if (!oor) {
                r = f32muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeFlonum((double)vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_F32VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = f32num(vvy, &oor);
            if (!oor) {
                r = f32muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeFlonum((double)vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of f32vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeFlonum(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeFlonum(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}
ScmObj Scm_F64VectorDotProd(ScmF64Vector *x, ScmObj y)
{
    int i, size = SCM_F64VECTOR_SIZE(x), oor;
    double r, vx, vy;
    ScmObj rr = SCM_MAKE_INT(0), vvy, vvx;
    
    r = 0;
    switch (arg2_check("f64vector-dot", SCM_OBJ(x), y, FALSE)) {
    case ARGTYPE_UVECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_F64VECTOR_ELEMENTS(x)[i];
            vy = SCM_F64VECTOR_ELEMENTS(y)[i];
            r = f64muladd(vx, vy, r, &rr);
        }
        break;
    case ARGTYPE_VECTOR:
        for (i=0; i<size; i++) {
            vx = SCM_F64VECTOR_ELEMENTS(x)[i];
            vvy = SCM_VECTOR_ELEMENTS(y)[i];
            vy = f64num(vvy, &oor);
            if (!oor) {
                r = f64muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeFlonum(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_LIST:
        for (i=0; i<size; i++) {
            vx = SCM_F64VECTOR_ELEMENTS(x)[i];
            vvy = SCM_CAR(y); y = SCM_CDR(y);
            vy = f64num(vvy, &oor);
            if (!oor) {
                r = f64muladd(vx, vy, r, &rr);
            } else {
                vvx = Scm_MakeFlonum(vx);
                rr = Scm_Add(rr, Scm_Mul(vvx, vvy));
            }
        }
        break;
    case ARGTYPE_CONST:
        Scm_Panic("something wrong");
    }

    /* r may contain a value bigger than the normal element value
       of f64vector, so it needs some care. */
    if (SCM_EQ(rr, SCM_MAKE_INT(0))) {
        rr = Scm_MakeFlonum(r);
    } else {
        ScmObj sr;
        sr = Scm_MakeFlonum(r);
        rr = Scm_Add(rr, sr);
    }
    return rr;
}

ScmObj Scm_S8VectorRangeCheck(ScmS8Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S8VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s8vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s8vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = s8unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = s8unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S8VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S8VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S8VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_S8VectorClamp(ScmS8Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S8VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s8vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s8vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = s8unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = s8unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S8VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S8VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S8VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_S8VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_S8VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_S8VectorClampX(ScmS8Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S8VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s8vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s8vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = s8unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = s8unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S8VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S8VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S8VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_S8VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_S8VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_U8VectorRangeCheck(ScmU8Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U8VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u8vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u8vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = u8unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = u8unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U8VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U8VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U8VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_U8VectorClamp(ScmU8Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U8VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u8vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u8vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = u8unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = u8unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U8VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U8VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U8VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_U8VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_U8VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_U8VectorClampX(ScmU8Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U8VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u8vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u8vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = u8unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = u8unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U8VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U8VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U8VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u8unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_U8VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_U8VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_S16VectorRangeCheck(ScmS16Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S16VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s16vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s16vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = s16unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = s16unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S16VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S16VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S16VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_S16VectorClamp(ScmS16Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S16VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s16vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s16vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = s16unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = s16unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S16VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S16VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S16VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_S16VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_S16VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_S16VectorClampX(ScmS16Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S16VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s16vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s16vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = s16unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = s16unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S16VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S16VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S16VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = s16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_S16VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_S16VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_U16VectorRangeCheck(ScmU16Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U16VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u16vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u16vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = u16unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = u16unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U16VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U16VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U16VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_U16VectorClamp(ScmU16Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U16VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u16vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u16vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = u16unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = u16unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U16VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U16VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U16VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_U16VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_U16VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_U16VectorClampX(ScmU16Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U16VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u16vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u16vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = u16unbox(min, SCM_CLAMP_BOTH);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = u16unbox(max, SCM_CLAMP_BOTH);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U16VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U16VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U16VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = u16unbox(vv, SCM_CLAMP_BOTH);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_U16VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_U16VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_S32VectorRangeCheck(ScmS32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s32vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s32vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetInteger32Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_S32VectorClamp(ScmS32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s32vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s32vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetInteger32Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_S32VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_S32VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_S32VectorClampX(ScmS32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s32vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s32vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetInteger32Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_S32VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_S32VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_U32VectorRangeCheck(ScmU32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    u_long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u32vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u32vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_U32VectorClamp(ScmU32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    u_long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u32vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u32vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_U32VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_U32VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_U32VectorClampX(ScmU32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    u_long val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u32vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u32vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU32Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_U32VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_U32VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_S64VectorRangeCheck(ScmS64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    ScmInt64 val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    SCM_SET_INT64_ZERO(minval);             /* make compiler happy */
    SCM_SET_INT64_ZERO(maxval);             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s64vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s64vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetInteger64Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && INT64LT(val, minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && INT64LT(maxval, val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_S64VectorClamp(ScmS64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    ScmInt64 val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    SCM_SET_INT64_ZERO(minval);             /* make compiler happy */
    SCM_SET_INT64_ZERO(maxval);             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s64vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s64vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetInteger64Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && INT64LT(val, minval)) {
            val = minval;
            SCM_S64VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && INT64LT(maxval, val)) {
            val = maxval;
            SCM_S64VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_S64VectorClampX(ScmS64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_S64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    ScmInt64 val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    SCM_SET_INT64_ZERO(minval);             /* make compiler happy */
    SCM_SET_INT64_ZERO(maxval);             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("s64vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("s64vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetInteger64Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_S64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_S64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_S64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetInteger64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && INT64LT(val, minval)) {
            val = minval;
            SCM_S64VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && INT64LT(maxval, val)) {
            val = maxval;
            SCM_S64VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_U64VectorRangeCheck(ScmU64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    ScmUInt64 val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    SCM_SET_INT64_ZERO(minval);             /* make compiler happy */
    SCM_SET_INT64_ZERO(maxval);             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u64vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u64vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && INT64LT(val, minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && INT64LT(maxval, val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_U64VectorClamp(ScmU64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    ScmUInt64 val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    SCM_SET_INT64_ZERO(minval);             /* make compiler happy */
    SCM_SET_INT64_ZERO(maxval);             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u64vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u64vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && INT64LT(val, minval)) {
            val = minval;
            SCM_U64VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && INT64LT(maxval, val)) {
            val = maxval;
            SCM_U64VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_U64VectorClampX(ScmU64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_U64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    ScmUInt64 val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    SCM_SET_INT64_ZERO(minval);             /* make compiler happy */
    SCM_SET_INT64_ZERO(maxval);             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("u64vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("u64vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(min, SCM_CLAMP_BOTH, NULL);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(max, SCM_CLAMP_BOTH, NULL);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_U64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_U64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_U64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetIntegerU64Clamp(vv, SCM_CLAMP_BOTH, NULL);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && INT64LT(val, minval)) {
            val = minval;
            SCM_U64VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && INT64LT(maxval, val)) {
            val = maxval;
            SCM_U64VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_F32VectorRangeCheck(ScmF32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_F32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    double val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("f32vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("f32vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetDouble(min);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetDouble(max);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_F32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_F32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_F32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_F32VectorClamp(ScmF32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_F32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    double val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("f32vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("f32vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetDouble(min);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetDouble(max);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_F32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_F32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_F32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_F32VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_F32VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_F32VectorClampX(ScmF32Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_F32VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    double val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("f32vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("f32vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetDouble(min);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetDouble(max);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_F32VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_F32VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_F32VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_F32VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_F32VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

ScmObj Scm_F64VectorRangeCheck(ScmF64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_F64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    double val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("f64vector-range-check", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("f64vector-range-check", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetDouble(min);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetDouble(max);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_F64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_F64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_F64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            return Scm_MakeInteger(i);
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            return Scm_MakeInteger(i);
        }
    }
    return SCM_FALSE;
}

ScmObj Scm_F64VectorClamp(ScmF64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_F64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    double val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ScmObj d = Scm_MakeUVector(Scm_ClassOf(SCM_OBJ(x)), SCM_UVECTOR_SIZE(x), SCM_UVECTOR_ELEMENTS(x));

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("f64vector-clamp", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("f64vector-clamp", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetDouble(min);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetDouble(max);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_F64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_F64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_F64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_F64VECTOR_ELEMENTS(d)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_F64VECTOR_ELEMENTS(d)[i] = val;
        }
    }
    return d;
}

ScmObj Scm_F64VectorClampX(ScmF64Vector *x, ScmObj min, ScmObj max)
{
    int i, size = SCM_F64VECTOR_SIZE(x);
    ArgType mintype, maxtype;
    double val, minval, maxval;
    int mindc = FALSE, maxdc = FALSE;         /* true if "don't care" */
    ScmObj vv;
    ;

    minval = 0;             /* make compiler happy */
    maxval = 0;             /* dittox */
        
    /* size check */
    if (SCM_FALSEP(min)) mintype = ARGTYPE_CONST;
    else mintype = arg2_check("f64vector-clamp!", SCM_OBJ(x), min, TRUE);

    if (SCM_FALSEP(max)) maxtype = ARGTYPE_CONST;
    else maxtype = arg2_check("f64vector-clamp!", SCM_OBJ(x), max, TRUE);

    if (mintype == ARGTYPE_CONST) {
        if ((mindc = SCM_FALSEP(min)) == FALSE) {
  minval = Scm_GetDouble(min);
};
    }
    if (maxtype == ARGTYPE_CONST) {
        if ((maxdc = SCM_FALSEP(max)) == FALSE) {
  maxval = Scm_GetDouble(max);
};
    }

    for (i=0; i<size; i++) {
        val = SCM_F64VECTOR_ELEMENTS(x)[i];
        switch (mintype) {
        case ARGTYPE_UVECTOR:
            minval = SCM_F64VECTOR_ELEMENTS(min)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(min)[i];
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(min); min = SCM_CDR(min);
            if ((mindc = SCM_FALSEP(vv)) == FALSE) {
  minval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }
        switch (maxtype) {
        case ARGTYPE_UVECTOR:
            maxval = SCM_F64VECTOR_ELEMENTS(max)[i]; break;
        case ARGTYPE_VECTOR:
            vv = SCM_VECTOR_ELEMENTS(max)[i];
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_LIST:
            vv = SCM_CAR(max); max = SCM_CDR(max);
            if ((maxdc = SCM_FALSEP(vv)) == FALSE) {
  maxval = Scm_GetDouble(vv);
};
            break;
        case ARGTYPE_CONST:
            /* alreadh handled above */
            break;
        }

        if (!mindc && (val < minval)) {
            val = minval;
            SCM_F64VECTOR_ELEMENTS(x)[i] = val;
        }
        if (!maxdc && (maxval < val)) {
            val = maxval;
            SCM_F64VECTOR_ELEMENTS(x)[i] = val;
        }
    }
    return SCM_OBJ(x);
}

static void s16vector_swapb(ScmS16Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    short *d = SCM_S16VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb16((void*)d);
    }
}

ScmObj Scm_S16VectorSwapBytes(ScmS16Vector *v)
{
    ScmObj d = Scm_S16VectorCopy(v, 0, -1);
    s16vector_swapb(SCM_S16VECTOR(d));
    return d;
}

ScmObj Scm_S16VectorSwapBytesX(ScmS16Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    s16vector_swapb(v);
    return SCM_OBJ(v);
}




static void u16vector_swapb(ScmU16Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    unsigned short *d = SCM_U16VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb16((void*)d);
    }
}

ScmObj Scm_U16VectorSwapBytes(ScmU16Vector *v)
{
    ScmObj d = Scm_U16VectorCopy(v, 0, -1);
    u16vector_swapb(SCM_U16VECTOR(d));
    return d;
}

ScmObj Scm_U16VectorSwapBytesX(ScmU16Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    u16vector_swapb(v);
    return SCM_OBJ(v);
}




static void s32vector_swapb(ScmS32Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    ScmInt32 *d = SCM_S32VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb32((void*)d);
    }
}

ScmObj Scm_S32VectorSwapBytes(ScmS32Vector *v)
{
    ScmObj d = Scm_S32VectorCopy(v, 0, -1);
    s32vector_swapb(SCM_S32VECTOR(d));
    return d;
}

ScmObj Scm_S32VectorSwapBytesX(ScmS32Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    s32vector_swapb(v);
    return SCM_OBJ(v);
}




static void u32vector_swapb(ScmU32Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    ScmUInt32 *d = SCM_U32VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb32((void*)d);
    }
}

ScmObj Scm_U32VectorSwapBytes(ScmU32Vector *v)
{
    ScmObj d = Scm_U32VectorCopy(v, 0, -1);
    u32vector_swapb(SCM_U32VECTOR(d));
    return d;
}

ScmObj Scm_U32VectorSwapBytesX(ScmU32Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    u32vector_swapb(v);
    return SCM_OBJ(v);
}




static void s64vector_swapb(ScmS64Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    ScmInt64 *d = SCM_S64VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb64((void*)d);
    }
}

ScmObj Scm_S64VectorSwapBytes(ScmS64Vector *v)
{
    ScmObj d = Scm_S64VectorCopy(v, 0, -1);
    s64vector_swapb(SCM_S64VECTOR(d));
    return d;
}

ScmObj Scm_S64VectorSwapBytesX(ScmS64Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    s64vector_swapb(v);
    return SCM_OBJ(v);
}




static void u64vector_swapb(ScmU64Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    ScmUInt64 *d = SCM_U64VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb64((void*)d);
    }
}

ScmObj Scm_U64VectorSwapBytes(ScmU64Vector *v)
{
    ScmObj d = Scm_U64VectorCopy(v, 0, -1);
    u64vector_swapb(SCM_U64VECTOR(d));
    return d;
}

ScmObj Scm_U64VectorSwapBytesX(ScmU64Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    u64vector_swapb(v);
    return SCM_OBJ(v);
}




static void f32vector_swapb(ScmF32Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    float *d = SCM_F32VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb32((void*)d);
    }
}

ScmObj Scm_F32VectorSwapBytes(ScmF32Vector *v)
{
    ScmObj d = Scm_F32VectorCopy(v, 0, -1);
    f32vector_swapb(SCM_F32VECTOR(d));
    return d;
}

ScmObj Scm_F32VectorSwapBytesX(ScmF32Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    f32vector_swapb(v);
    return SCM_OBJ(v);
}




static void f64vector_swapb(ScmF64Vector *v)
{
    int i, len = SCM_UVECTOR_SIZE(v);
    double *d = SCM_F64VECTOR_ELEMENTS(v);
    for (i=0; i<len; i++, d++) {
        swapb64((void*)d);
    }
}

ScmObj Scm_F64VectorSwapBytes(ScmF64Vector *v)
{
    ScmObj d = Scm_F64VectorCopy(v, 0, -1);
    f64vector_swapb(SCM_F64VECTOR(d));
    return d;
}

ScmObj Scm_F64VectorSwapBytesX(ScmF64Vector *v)
{
    SCM_UVECTOR_CHECK_MUTABLE(v);
    f64vector_swapb(v);
    return SCM_OBJ(v);
}




/*==============================================================
 * Some generic functions
 */

/*
 * Generic copy
 */
ScmObj Scm_UVectorCopy(ScmUVector *v, int start, int end)
{
    switch (uvector_index(Scm_ClassOf(SCM_OBJ(v)))) {
    case 0: return Scm_S8VectorCopy(SCM_S8VECTOR(v), start, end);
    case 1: return Scm_U8VectorCopy(SCM_U8VECTOR(v), start, end);
    case 2: return Scm_S16VectorCopy(SCM_S16VECTOR(v), start, end);
    case 3: return Scm_U16VectorCopy(SCM_U16VECTOR(v), start, end);
    case 4: return Scm_S32VectorCopy(SCM_S32VECTOR(v), start, end);
    case 5: return Scm_U32VectorCopy(SCM_U32VECTOR(v), start, end);
    case 6: return Scm_S64VectorCopy(SCM_S64VECTOR(v), start, end);
    case 7: return Scm_U64VectorCopy(SCM_U64VECTOR(v), start, end);
    case 8: return Scm_F32VectorCopy(SCM_F32VECTOR(v), start, end);
    case 9: return Scm_F64VectorCopy(SCM_F64VECTOR(v), start, end);
    default: Scm_Error("uniform vector required, but got %S", v);
        return SCM_UNDEFINED;
    }
}

/*
 * Generic swapb
 */
ScmObj Scm_UVectorSwapBytes(ScmUVector *v)
{
    switch (uvector_index(Scm_ClassOf(SCM_OBJ(v)))) {
    case 0: return SCM_OBJ(v);
    case 1: return SCM_OBJ(v);
    case 2: return Scm_S16VectorSwapBytes(SCM_S16VECTOR(v));
    case 3: return Scm_U16VectorSwapBytes(SCM_U16VECTOR(v));
    case 4: return Scm_S32VectorSwapBytes(SCM_S32VECTOR(v));
    case 5: return Scm_U32VectorSwapBytes(SCM_U32VECTOR(v));
    case 6: return Scm_S64VectorSwapBytes(SCM_S64VECTOR(v));
    case 7: return Scm_U64VectorSwapBytes(SCM_U64VECTOR(v));
    case 8: return Scm_F32VectorSwapBytes(SCM_F32VECTOR(v));
    case 9: return Scm_F64VectorSwapBytes(SCM_F64VECTOR(v));
    default: Scm_Error("uniform vector required, but got %S", v);
        return SCM_UNDEFINED;
    }
}

ScmObj Scm_UVectorSwapBytesX(ScmUVector *v)
{
    switch (uvector_index(Scm_ClassOf(SCM_OBJ(v)))) {
    case 0: return SCM_OBJ(v);
    case 1: return SCM_OBJ(v);
    case 2: return Scm_S16VectorSwapBytesX(SCM_S16VECTOR(v));
    case 3: return Scm_U16VectorSwapBytesX(SCM_U16VECTOR(v));
    case 4: return Scm_S32VectorSwapBytesX(SCM_S32VECTOR(v));
    case 5: return Scm_U32VectorSwapBytesX(SCM_U32VECTOR(v));
    case 6: return Scm_S64VectorSwapBytesX(SCM_S64VECTOR(v));
    case 7: return Scm_U64VectorSwapBytesX(SCM_U64VECTOR(v));
    case 8: return Scm_F32VectorSwapBytesX(SCM_F32VECTOR(v));
    case 9: return Scm_F64VectorSwapBytesX(SCM_F64VECTOR(v));
    default: Scm_Error("uniform vector required, but got %S", v);
        return SCM_UNDEFINED;
    }
}

/*
 * Block I/O
 */
static void endian_check(ScmObj endian)
{
    if (!SCM_FALSEP(endian)
        && !SCM_EQ(endian, SCM_SYM_BIG_ENDIAN)
        && !SCM_EQ(endian, SCM_SYM_LITTLE_ENDIAN)) {
        Scm_Error("endian argument must be either #f, big-endian or little-endian, but got: %S", endian);
    }
}

ScmObj Scm_ReadBlockX(ScmUVector *v, ScmPort *port,
                      int start, int end, ScmObj endian)
{
    int len = SCM_UVECTOR_SIZE(v), eltsize, r;

    SCM_CHECK_START_END(start, end, len);
    SCM_UVECTOR_CHECK_MUTABLE(v);
    endian_check(endian);

    eltsize = Scm_UVectorElementSize(Scm_ClassOf(SCM_OBJ(v)));
    SCM_ASSERT(eltsize >= 1);
    r = Scm_Getz((char*)v->elements + start*eltsize,
                 (end-start)*eltsize, port);
    if (r == EOF) SCM_RETURN(SCM_EOF);
#ifdef WORDS_BIGENDIAN
    if (SCM_EQ(endian, SCM_SYM_LITTLE_ENDIAN)) {
        Scm_UVectorSwapBytesX(SCM_UVECTOR(v));
    }
#else /*!WORDS_BIGENDIAN*/
    if (SCM_EQ(endian, SCM_SYM_BIG_ENDIAN)) {
        Scm_UVectorSwapBytesX(SCM_UVECTOR(v));
    }
#endif
    SCM_RETURN(Scm_MakeInteger((r+eltsize-1)/eltsize));
}

ScmObj Scm_WriteBlock(ScmUVector *v, ScmPort *port,
                      int start, int end, ScmObj endian)
{
    int len = SCM_UVECTOR_SIZE(v), eltsize, swapb = FALSE;
    SCM_CHECK_START_END(start, end, len);
    endian_check(endian);
    
    eltsize = Scm_UVectorElementSize(Scm_ClassOf(SCM_OBJ(v)));
    SCM_ASSERT(eltsize >= 1);
#ifdef WORDS_BIGENDIAN
    swapb = SCM_EQ(endian, SCM_SYM_LITTLE_ENDIAN);
#else /*!WORDS_BIGENDIAN*/
    swapb = SCM_EQ(endian, SCM_SYM_BIG_ENDIAN);
#endif
    if (!swapb || eltsize == 1) {
        Scm_Putz((const char*)v->elements + start*eltsize,
                 (end-start)*eltsize, port);
    } else {
        /* ugly */
        int i;
        
        switch (eltsize) {
        case 2: {
            union {
                unsigned short n;
                unsigned char c[2];
            } d;
            for (i=start; i<end; i++) {
                d.n = ((unsigned short*)v->elements)[i];
                swapb16(&d.n);
                Scm_Putz((const char*)d.c, 2, port);
            }
        }
        case 4: {
            union {
                ScmUInt32 n;
                unsigned char c[4];
            } d;
            for (i=start; i<end; i++) {
                d.n = ((ScmUInt32*)v->elements)[i];
                swapb32(&d.n);
                Scm_Putz((const char*)d.c, 4, port);
            }
        }
        case 8:{
            union {
                ScmUInt64 n;
                unsigned char c[8];
            } d;
            for (i=start; i<end; i++) {
                d.n = ((ScmUInt64*)v->elements)[i];
                swapb64(&d.n);
                Scm_Putz((const char*)d.c, 8, port);
            }
        }
        }
    }
    SCM_RETURN(SCM_UNDEFINED);
}

